import pandas as pd

from lac_validator.rule_engine import rule_definition
from lac_validator.rules.rule_utils import field_different_from_previous


@rule_definition(
    code="503A",
    message="The reason for new episode in the first episode does not match open episode at end of last year.",
    affected_fields=["RNE"],
    tables=["Episodes", "Episodes_last"],
)
def validate(dfs):
    return field_different_from_previous(dfs, field="RNE")


def test_validate():
    fake_epi = pd.DataFrame(
        [
            {
                "CHILD": "111",
                "DEC": pd.NA,
                "DECOM": "01/06/2020",
                "RNE": "L",
            },  # 0  Min, Fails
            {"CHILD": "111", "DEC": pd.NA, "DECOM": "05/06/2020", "RNE": "L"},  # 1
            {"CHILD": "111", "DEC": pd.NA, "DECOM": "06/06/2020", "RNE": "L"},  # 2
            {
                "CHILD": "123",
                "DEC": pd.NA,
                "DECOM": "08/06/2020",
                "RNE": "L",
            },  # 3  Min, Fails
            {"CHILD": "222", "DEC": pd.NA, "DECOM": "05/06/2020", "RNE": "L"},  # 4  Min
            {
                "CHILD": "333",
                "DEC": pd.NA,
                "DECOM": "06/06/2020",
                "RNE": "L",
            },  # 5  Min, Fails
            {"CHILD": "333", "DEC": pd.NA, "DECOM": "07/06/2020", "RNE": "L"},  # 6
            {"CHILD": "444", "DEC": pd.NA, "DECOM": "08/06/2020", "RNE": "L"},  # 7  Min
        ]
    )

    fake_epi_last = pd.DataFrame(
        [
            {
                "CHILD": "111",
                "DEC": pd.NA,
                "DECOM": "01/06/2020",
                "RNE": "S",
            },  # Max Different RNE
            {
                "CHILD": "123",
                "DEC": pd.NA,
                "DECOM": "08/06/2020",
                "RNE": "P",
            },  # Max Different RNE
            {
                "CHILD": "222",
                "DEC": pd.NA,
                "DECOM": "05/06/2020",
                "RNE": "L",
            },  # Max Same
            {
                "CHILD": "333",
                "DEC": pd.NA,
                "DECOM": "06/06/2020",
                "RNE": "R",
            },  # Max Different RNE
            {"CHILD": "444", "DEC": pd.NA, "DECOM": "08/06/2020", "RNE": "L"},
            {"CHILD": "444", "DEC": pd.NA, "DECOM": "09/06/2020", "RNE": "L"},
            {
                "CHILD": "444",
                "DEC": pd.NA,
                "DECOM": "19/06/2020",
                "RNE": "L",
            },  # Max different date so passes
        ]
    )
    fake_epi["DECOM"] = pd.to_datetime(
        fake_epi["DECOM"], format="%d/%m/%Y", errors="coerce"
    )
    fake_epi_last["DECOM"] = pd.to_datetime(
        fake_epi_last["DECOM"], format="%d/%m/%Y", errors="coerce"
    )
    fake_epi_last["DEC"] = pd.to_datetime(
        fake_epi_last["DEC"], format="%d/%m/%Y", errors="coerce"
    )

    # For the test, this doesn't affect anything except how the data is handled
    # by the field_different function, so it doesn't matter that the years
    # in the test dfs are not accurate to this collection year
    # As seen in conftest collection year is read in as XXXX/XX and stored in metadata
    # as the first four characters using [:4] in _process_metadata in datastore```
    metadata = {"collectionYear": "2022"}

    fake_dfs = {
        "Episodes": fake_epi,
        "Episodes_last": fake_epi_last,
        "metadata": metadata,
    }

    result = validate(fake_dfs)

    assert result == {"Episodes": [0, 3, 5]}



import importlib
from typing import Iterable

from lac_validator.rule_engine import RuleDefinition, YearConfig


def extract_validator_functions(
    file_paths: Iterable, marker: str = "rule"
) -> dict[str, RuleDefinition]:
    """
    :param list file_paths: list of file paths to extract functions from.
    :param str marker: marker to identify functions to extract.

    :return: functions extracted from files.
    :rtype: dict
    """
    validator_funcs: dict[str, RuleDefinition] = {}
    for path in file_paths:
        if path.stem == "__init__":
            continue
        rule_content = importlib.import_module(
            f"lac_validator.rules.{path.parent.stem}.{path.stem}"
        )
        validator_func = {
            element.rule.code: element.rule
            for _, element in vars(rule_content).items()
            if hasattr(element, marker)
        }
        duplicate_funcs = set(validator_func.keys()) & set(validator_funcs.keys())
        if duplicate_funcs:
            raise ValueError(f"Rule with code {duplicate_funcs} already exists")
        validator_funcs.update(validator_func)
    return validator_funcs


def update_validator_functions(
    prev_validator_funcs, this_year_config: YearConfig
) -> dict:
    """
    :param dict prev_registry: previous year's registry.
    :param dict this_year_config: codes of rules that have been added or deleted.

    :return: valid validator functions according to config.
    :rtype: dict
    """
    # Rules present in both will be updated to this year's version. rules present in only this year will be added.
    updated_validator_funcs = prev_validator_funcs | this_year_config.added_or_modified
    # delete rules by their rules codes, if specified.
    for deleted_rule in this_year_config.deleted:
        del updated_validator_funcs[deleted_rule]
    return updated_validator_funcs


def get_year_ruleset(collection_year: str) -> dict[str, RuleDefinition]:
    """
    Gets the registry of validation rules for the year specified in the metadata.
    """
    # for example, convert "2023" to "lac2022_23"
    ruleset = f"lac{int(collection_year)-1}_{collection_year[2:4]}"

    module = importlib.import_module(f"lac_validator.rules.{ruleset}")
    registry = getattr(module, "registry")

    return registry



import pandas as pd

from lac_validator.rule_engine import rule_definition


@rule_definition(
    code="502",
    message="Last year's record ended with an open episode. The date on which that episode started does not match the start date of the first episode on this year’s record.",
    affected_fields=["DECOM"],
    tables=["Episodes", "Episodes_last"],
)
def validate(dfs):
    if "Episodes" not in dfs or "Episodes_last" not in dfs:
        return {}
    else:
        epi = dfs["Episodes"]
        epi_last = dfs["Episodes_last"]
        epi = epi.reset_index()

        epi["DECOM"] = pd.to_datetime(epi["DECOM"], format="%d/%m/%Y", errors="coerce")
        epi_last["DECOM"] = pd.to_datetime(
            epi_last["DECOM"], format="%d/%m/%Y", errors="coerce"
        )

        epi_last_no_dec = epi_last[epi_last["DEC"].isna()]

        epi_min_decoms_index = (
            epi[["CHILD", "DECOM"]].groupby(["CHILD"])["DECOM"].idxmin()
        )

        epi_min_decom_df = epi.loc[epi_min_decoms_index, :]

        merged_episodes = epi_min_decom_df.merge(
            epi_last_no_dec, on="CHILD", how="inner"
        )
        error_cohort = merged_episodes[
            merged_episodes["DECOM_x"] != merged_episodes["DECOM_y"]
        ]

        error_list = error_cohort["index"].to_list()
        error_list = list(set(error_list))
        error_list.sort()

        return {"Episodes": error_list}


def test_validate():
    import pandas as pd

    fake_epi = pd.DataFrame(
        [
            {"CHILD": "111", "DECOM": "01/06/2020"},  # 0  Min   Fails
            {"CHILD": "111", "DECOM": "05/06/2020"},  # 1
            {"CHILD": "111", "DECOM": "06/06/2020"},  # 2
            {"CHILD": "123", "DECOM": "08/06/2020"},  # 3  Min
            {"CHILD": "222", "DECOM": "05/06/2020"},  # 4   Min   Fails
            {"CHILD": "333", "DECOM": "06/06/2020"},  # 5  Min
            {"CHILD": "333", "DECOM": "07/06/2020"},  # 6
            {"CHILD": "444", "DECOM": "08/06/2020"},  # 7  Min   Fails
            {"CHILD": "444", "DECOM": "09/06/2020"},  # 8
            {"CHILD": "444", "DECOM": "15/06/2020"},  # 9
            {"CHILD": "555", "DECOM": "15/06/2020"},  # 10
        ]
    )

    fake_epi_last = pd.DataFrame(
        [
            {"CHILD": "111", "DECOM": "05/06/2020", "DEC": pd.NA},
            {"CHILD": "123", "DECOM": "08/06/2020", "DEC": pd.NA},
            {"CHILD": "222", "DECOM": "09/06/2020", "DEC": pd.NA},
            {"CHILD": "333", "DECOM": "06/06/2020", "DEC": "05/06/2020"},
            {"CHILD": "333", "DECOM": "07/06/2020", "DEC": "05/06/2020"},
            {"CHILD": "444", "DECOM": "08/06/2020", "DEC": "05/06/2020"},
            {"CHILD": "444", "DECOM": "09/06/2020", "DEC": "05/06/2020"},
            {"CHILD": "444", "DECOM": "19/06/2020", "DEC": pd.NA},
        ]
    )

    fake_dfs = {"Episodes": fake_epi, "Episodes_last": fake_epi_last}

    result = validate(fake_dfs)

    assert result == {"Episodes": [0, 4, 7]}



import pandas as pd

from lac_validator.rule_engine import rule_definition


@rule_definition(
    code="426",
    message="A child receiving respite care cannot be recorded under a legal status of V3 and V4 in "
    + "the same year.",
    affected_fields=["LS"],
    tables=["Episodes"],
)
def validate(dfs):
    if "Episodes" not in dfs:
        return {}
    else:
        epi = dfs["Episodes"]
        epi_v3 = epi[epi["LS"] == "V3"]
        epi_v4 = epi[epi["LS"] == "V4"]

        m_coh = epi_v3.merge(epi_v4, on="CHILD", how="inner")
        err_child = m_coh["CHILD"].unique().tolist()

        err_l1 = epi_v3[epi_v3["CHILD"].isin(err_child)].index.tolist()
        err_l2 = epi_v4[epi_v4["CHILD"].isin(err_child)].index.tolist()

        err_list = err_l1 + err_l2
        err_list.sort()
        return {"Episodes": err_list}


def test_validate():
    import pandas as pd

    fake_data = pd.DataFrame(
        [
            {"CHILD": "111", "LS": "V3"},  # 0
            {"CHILD": "111", "LS": "V3"},  # 1
            {"CHILD": "222", "LS": "V3"},  # 2 Fail
            {"CHILD": "222", "LS": "V4"},  # 3 Fail
            {"CHILD": "222", "LS": "X1"},  # 4
            {"CHILD": "444", "LS": "V3"},  # 5
            {"CHILD": "555", "LS": "V4"},  # 6 Fail
            {"CHILD": "555", "LS": "V3"},  # 7 Fail
            {"CHILD": "555", "LS": "V3"},  # 8 Fail
        ]
    )

    fake_dfs = {"Episodes": fake_data}

    result = validate(fake_dfs)

    assert result == {"Episodes": [2, 3, 6, 7, 8]}


# EPI is an undocumented rule in the DFE portal. It checks whether each Child ID in the Header file exists in either the Episodes or OC3 file.

# Testing rule with only Header and Episodes provided.  Rule will also function with only Header provided as per DFE portal.



import pandas as pd

from lac_validator.rule_engine import rule_definition


@rule_definition(
    code="118",
    message="Date of decision that a child should no longer be placed for adoption is before the current collection year or before the date the child started to be looked after.",
    affected_fields=["DECOM", "DECOM", "LS"],
    tables=["Episodes", "PlacedAdoption"],
)
def validate(dfs):
    if ("PlacedAdoption" not in dfs) or ("Episodes" not in dfs):
        return {}
    else:
        placed_adoption = dfs["PlacedAdoption"]
        episodes = dfs["Episodes"]
        collection_start = dfs["metadata"]["collection_start"]
        code_list = ["V3", "V4"]

        # datetime
        episodes["DECOM"] = pd.to_datetime(
            episodes["DECOM"], format="%d/%m/%Y", errors="coerce"
        )
        placed_adoption["DATE_PLACED_CEASED"] = pd.to_datetime(
            placed_adoption["DATE_PLACED_CEASED"],
            format="%d/%m/%Y",
            errors="coerce",
        )
        collection_start = pd.to_datetime(
            collection_start, format="%d/%m/%Y", errors="coerce"
        )

        # <DECOM> of the earliest episode with an <LS> not = 'V3' or 'V4'
        filter_by_ls = episodes[~(episodes["LS"].isin(code_list))]
        earliest_episode_idxs = filter_by_ls.groupby("CHILD")["DECOM"].idxmin()
        earliest_episodes = episodes[episodes.index.isin(earliest_episode_idxs)]

        # prepare to merge
        placed_adoption.reset_index(inplace=True)
        earliest_episodes.reset_index(inplace=True)

        # merge
        merged = earliest_episodes.merge(
            placed_adoption, on="CHILD", how="left", suffixes=["_eps", "_pa"]
        )

        # drop rows where DATE_PLACED_CEASED is not provided
        merged = merged.dropna(subset=["DATE_PLACED_CEASED"])
        # If provided <DATE_PLACED_CEASED> must not be prior to <COLLECTION_START_DATE> or <DECOM> of the earliest episode with an <LS> not = 'V3' or 'V4'
        mask = (merged["DATE_PLACED_CEASED"] < merged["DECOM"]) | (
            merged["DATE_PLACED_CEASED"] < collection_start
        )
        # error locations
        pa_error_locs = merged.loc[mask, "index_pa"]
        eps_error_locs = merged.loc[mask, "index_eps"]
        return {
            "Episodes": eps_error_locs.tolist(),
            "PlacedAdoption": pa_error_locs.unique().tolist(),
        }


def test_validate():
    import pandas as pd

    fake_placed_adoption = pd.DataFrame(
        {
            "DATE_PLACED_CEASED": ["08/03/2020", "22/06/2020", "13/10/2022", pd.NA],
            "CHILD": ["101", "102", "103", "104"],
        }
    )
    fake_data_episodes = pd.DataFrame(
        [
            {
                "CHILD": "101",
                "LS": "L1",
                "DECOM": "01/01/2019",
            },  # 0 Fail DATE_PLACED_CEASED is before collection_start
            {
                "CHILD": "102",
                "LS": "V3",
                "DECOM": "01/01/2022",
            },  # 1 skip fail because LS is V3
            {"CHILD": "102", "LS": "X0", "DECOM": "20/12/2020"},  # 2 fail
            {"CHILD": "102", "LS": "L1", "DECOM": "03/01/2021"},  # 3
            {"CHILD": "102", "LS": "L1", "DECOM": "03/04/2022"},  # 4
            {"CHILD": "103", "LS": "X2", "DECOM": "01/01/2019"},  # 5 pass
            {
                "CHILD": "104",
                "LS": "L1",
                "DECOM": "01/01/2019",
            },  # 6 drop.na drops this child
        ]
    )
    metadata = {
        "collection_start": "01/04/2020",
    }
    fake_dfs = {
        "Episodes": fake_data_episodes,
        "PlacedAdoption": fake_placed_adoption,
        "metadata": metadata,
    }

    result = validate(fake_dfs)
    assert result == {"Episodes": [0, 2], "PlacedAdoption": [0, 1]}



import pandas as pd

from lac_validator.rule_engine import rule_definition


@rule_definition(
    code="386",
    message="Reason episode ceased is adopted but child has reached age 18.",
    affected_fields=["REC"],
    tables=["Header", "Episodes"],
)
def validate(dfs):
    if "Header" not in dfs:
        return {}
    if "Episodes" not in dfs:
        return {}
    else:
        header = dfs["Header"]
        episodes = dfs["Episodes"]

        header["DOB"] = pd.to_datetime(
            header["DOB"], format="%d/%m/%Y", errors="coerce"
        )
        episodes["DEC"] = pd.to_datetime(
            episodes["DEC"], format="%d/%m/%Y", errors="coerce"
        )
        header["DOB18"] = header["DOB"] + pd.DateOffset(years=18)

        episodes_merged = (
            episodes.reset_index()
            .merge(
                header,
                how="left",
                on=["CHILD"],
                suffixes=("", "_header"),
                indicator=True,
            )
            .set_index("index")
            .dropna(subset=["DOB18", "DEC"])
        )

        ceased_adopted = (
            episodes_merged["REC"].str.upper().astype(str).isin(["E11", "E12"])
        )
        ceased_under_18 = episodes_merged["DOB18"] > episodes_merged["DEC"]

        error_mask = ceased_adopted & ~ceased_under_18

        error_locations = episodes_merged.index[error_mask]

        return {"Episodes": error_locations.to_list()}


def test_validate():
    import pandas as pd

    fake_data = pd.DataFrame(
        {
            "CHILD": ["101", "102", "101", "102", "103", "104"],
            "REC": ["E11", "E12", "X1", pd.NA, "E11", "E11"],
            "DEC": [
                "15/03/2021",
                "17/06/2020",
                "20/03/2020",
                pd.NA,
                "23/08/2020",
                pd.NA,
            ],
        }
    )

    fake_data_child = pd.DataFrame(
        {
            "CHILD": ["101", "102", "103", "104"],
            "DOB": ["16/03/2020", "23/09/2002", "31/12/2000", "01/02/2003"],
        }
    )

    fake_dfs = {"Episodes": fake_data, "Header": fake_data_child}

    result = validate(fake_dfs)

    assert result == {"Episodes": [4]}



import pandas as pd

from lac_validator.rule_engine import rule_definition


@rule_definition(
    code="168",
    message="Unique Pupil Number (UPN) is not valid. If unknown, default codes should be UN1, UN2, UN3, UN4 or UN5.",
    affected_fields=["UPN"],
    tables=["Header"],
)
def validate(dfs):
    if "Header" not in dfs:
        return {}
    else:
        df = dfs["Header"]
        mask = df["UPN"].str.match(
            r"(^((?![IOS])[A-Z]){1}(\d{12}|\d{11}[A-Z]{1})$)|^(UN[1-5])$", na=False
        )
        mask = ~mask & df["UPN"].notnull()
        return {"Header": df.index[mask].tolist()}


def test_validate():
    import pandas as pd

    fake_data = pd.DataFrame(
        {
            "UPN": [
                "UN3",
                "E07295962325C1556",
                "UN5",
                "UN7",
                "UPN3sDSG",
                "X06805040829A",
                "5035247309594",
                pd.NA,
                "L086819786126",
                "K06014812931",
                "J000947841350156",
                "M0940709",
                "I072272729588",
                "N075491517151",
                "Z041674136429",
                "E043016488226",
                "S074885779408",
            ],
        }
    )

    fake_dfs = {"Header": fake_data}

    result = validate(fake_dfs)

    assert result == {"Header": [1, 3, 4, 6, 9, 10, 11, 12, 16]}



import pandas as pd

from lac_validator.rule_engine import rule_definition


@rule_definition(
    code="631",
    message="Previous permanence option not a valid value.",
    affected_fields=["PREV_PERM"],
    tables=["PrevPerm"],
)
def validate(dfs):
    if "PrevPerm" not in dfs:
        return {}

    previous_permanence = dfs["PrevPerm"]
    code_list = ["P1", "P2", "P3", "P4", "Z1"]

    mask = (
        previous_permanence["PREV_PERM"].isin(code_list)
        | previous_permanence["PREV_PERM"].isna()
    )

    validation_error_mask = ~mask
    validation_error_locations = previous_permanence.index[validation_error_mask]

    return {"PrevPerm": validation_error_locations.tolist()}


def test_validate():
    import pandas as pd

    fake_data = pd.DataFrame(
        {
            "PREV_PERM": ["P0", "P1", "p1", "", pd.NA],
        }
    )

    fake_dfs = {"PrevPerm": fake_data}

    result = validate(fake_dfs)

    assert result == {"PrevPerm": [0, 2, 3]}



import pandas as pd

from lac_validator.rule_engine import rule_definition


@rule_definition(
    code="554",
    message="Date of decision that the child should be placed for adoption should be on or prior to the date that the placement order was granted. [NOTE: This rule may result in false positives or false negatives if relevant episodes are in previous years; please check carefully!]",
    affected_fields=["DATE_PLACED", "DECOM", "LS"],
    tables=["Episodes", "PlacedAdoption"],
)
def validate(dfs):
    if "Episodes" not in dfs or "PlacedAdoption" not in dfs:
        return {}
    else:
        episodes = dfs["Episodes"]
        placed_adoption = dfs["PlacedAdoption"]

        # convert dates from string format to datetime format.
        episodes["DECOM"] = pd.to_datetime(
            episodes["DECOM"], format="%d/%m/%Y", errors="coerce"
        )
        placed_adoption["DATE_PLACED"] = pd.to_datetime(
            placed_adoption["DATE_PLACED"], format="%d/%m/%Y", errors="coerce"
        )

        # Keep original index values as a column
        episodes["eps_index"] = episodes.index
        placed_adoption["pa_index"] = placed_adoption.index

        # select the first episodes with LS==E1
        e1_episodes = episodes.loc[episodes["LS"] == "E1"]
        first_e1_eps = e1_episodes.loc[e1_episodes.groupby("CHILD")["DECOM"].idxmin()]

        # merge
        merged = first_e1_eps.merge(placed_adoption, on="CHILD", how="left")

        # Where <LS> = 'E1' <DATE_PLACED> should be <= <DECOM> of first episode in <PERIOD_OF_CARE> with <LS> = 'E1'
        mask = merged["DATE_PLACED"] > merged["DECOM"]

        # error locations
        eps_error_locs = merged.loc[mask, "eps_index"]
        pa_error_locs = merged.loc[mask, "pa_index"]

        return {
            "Episodes": eps_error_locs.tolist(),
            "PlacedAdoption": pa_error_locs.unique().tolist(),
        }


def test_validate():
    import pandas as pd

    fake_placed_adoption = pd.DataFrame(
        [
            {"CHILD": 101, "DATE_PLACED": "26/05/2022"},
            # 0
            {"CHILD": 102, "DATE_PLACED": pd.NA},
            # 1
            {
                "CHILD": 103,
                "DATE_PLACED": "26/05/2020",
            },  # 2
            {"CHILD": 104, "DATE_PLACED": "22/01/2020"},
            # 3
            {
                "CHILD": 105,
                "DATE_PLACED": "11/01/2020",
            },  # 4
        ]
    )
    fake_data_episodes = pd.DataFrame(
        [
            # first E1 episode is first episode
            {
                "CHILD": 101,
                "DECOM": "01/01/2020",
                "LS": "E1",
            },  # 0 fail DATE_PLACED>DECOM
            {
                "CHILD": 101,
                "DECOM": "11/01/2020",
                "LS": "T1",
            },  # 1
            {
                "CHILD": 101,
                "DECOM": "22/01/2020",
                "LS": "A3",
            },  # 2
            # no E1 episode
            {
                "CHILD": 102,
                "DECOM": "11/01/2020",
                "LS": "A4",
            },  # 4
            {
                "CHILD": 102,
                "DECOM": "01/01/2020",
                "LS": "U1",
            },  # 3
            # first E1 episode is not first episode
            {
                "CHILD": 103,
                "DECOM": "01/01/2020",
                "LS": "U1",
            },  # 5
            {
                "CHILD": 103,
                "DECOM": "22/07/2020",
                "LS": "E1",
            },  # 6 pass
            {
                "CHILD": 103,
                "DECOM": "11/01/2020",
                "LS": "T2",
            },  # 7
            {
                "CHILD": 104,
                "DECOM": "01/02/2016",
                "LS": "E1",
            },  # 8 fail DATE_PLACED>DECOM
            {
                "CHILD": 104,
                "DECOM": "11/01/2020",
                "LS": "X1",
            },  # 9
            {
                "CHILD": 104,
                "DECOM": "01/01/2020",
                "LS": "A3",
            },  # 10
            {
                "CHILD": 105,
                "DECOM": "11/01/2020",
                "LS": "E1",
            },  # 11 pass DATE_PLACED equals DECOM
            {
                "CHILD": 105,
                "DECOM": "01/01/2020",
                "LS": pd.NA,
            },  # 12
        ]
    )
    fake_dfs = {"Episodes": fake_data_episodes, "PlacedAdoption": fake_placed_adoption}

    result = validate(fake_dfs)
    assert result == {"Episodes": [0, 8], "PlacedAdoption": [0, 3]}



import pandas as pd

from lac_validator.rule_engine import rule_definition


@rule_definition(
    code="441",
    message="Participation method indicates child was 4 years old or over at the time of the review, but the date of birth and review date indicates the child was under 4 years old.",
    affected_fields=["DOB", "REVIEW", "REVIEW_CODE"],
    tables=["Reviews"],
)
def validate(dfs):
    if "Reviews" not in dfs:
        return {}
    else:
        reviews = dfs["Reviews"]
        reviews["DOB"] = pd.to_datetime(
            reviews["DOB"], format="%d/%m/%Y", errors="coerce"
        )
        reviews["REVIEW"] = pd.to_datetime(
            reviews["REVIEW"], format="%d/%m/%Y", errors="coerce"
        )
        reviews = reviews.dropna(subset=["REVIEW", "DOB"])

        mask = reviews["REVIEW_CODE"].isin(
            ["PN1", "PN2", "PN3", "PN4", "PN5", "PN6", "PN7"]
        ) & (reviews["REVIEW"] < reviews["DOB"] + pd.offsets.DateOffset(years=4))

        validation_error_mask = mask

        validation_error_locations = reviews.index[validation_error_mask]

        return {"Reviews": validation_error_locations.tolist()}


def test_validate():
    import pandas as pd

    fake_data = pd.DataFrame(
        {
            "DOB": [
                "01/01/2004",
                "01/01/2006",
                "01/01/2007",
                "01/01/2008",
                "01/01/2010",
                "01/01/2012",
                "01/01/2014",
                "01/06/2014",
            ],
            "REVIEW": [
                "31/12/2008",
                "01/01/2012",
                pd.NA,
                "01/01/2009",
                "01/01/2012",
                "01/01/2017",
                "01/01/2021",
                "01/01/2015",
            ],
            "REVIEW_CODE": ["PN1", "PN2", "PN3", "PN4", "PN5", "PN6", "PN7", "PN0"],
        }
    )

    fake_dfs = {"Reviews": fake_data}

    result = validate(fake_dfs)

    assert result == {"Reviews": [3, 4]}



import pandas as pd

from lac_validator.rule_engine import rule_definition


@rule_definition(
    code="390",
    message="Reason episode ceased is adopted but child has not been previously placed for adoption.",
    affected_fields=["PLACE", "REC"],
    tables=["Episodes"],
)
def validate(dfs):
    if "Episodes" not in dfs:
        return {}
    else:
        episodes = dfs["Episodes"]
        # If <REC> = 'E11' or 'E12' then <PL> must be one of 'A3', 'A4', 'A5' or 'A6'
        mask = ((episodes["REC"] == "E11") | (episodes["REC"] == "E12")) & ~(
            (episodes["PLACE"] == "A3")
            | (episodes["PLACE"] == "A4")
            | (episodes["PLACE"] == "A5")
            | (episodes["PLACE"] == "A6")
        )
        error_locations = episodes.index[mask]
        return {"Episodes": error_locations.to_list()}


def test_validate():
    import pandas as pd

    fake_data = pd.DataFrame(
        [
            {"CHILD": "11", "PLACE": "P1", "LS": "x", "REC": "E11"},  # 0
            {"CHILD": "202", "PLACE": "x", "LS": "D1", "REC": "x"},  # 1
            {"CHILD": "3003", "PLACE": "A4", "LS": "D1", "REC": "E12"},  # 2
            {"CHILD": "40004", "PLACE": "P1", "LS": "V2", "REC": "E12"},  # 3
            {"CHILD": "5005", "PLACE": "A5", "LS": "x", "REC": "E11"},  # 4
            {"CHILD": "606", "PLACE": "A6", "LS": "V2", "REC": "x"},  # 5
            {"CHILD": "77", "PLACE": "x", "LS": "x", "REC": "x"},  # 6
        ]
    )
    fake_dfs = {"Episodes": fake_data}

    result = validate(fake_dfs)
    assert result == {"Episodes": [0, 3]}



import pandas as pd

from lac_validator.rule_engine import rule_definition


@rule_definition(
    code="224",
    message="The Ofsted Unique reference number (URN) provided for the child's placement does not match the placement provider recorded.",
    affected_fields=["PLACE_PROVIDER"],
    tables=["Episodes", "Provider Info"],
)
def validate(dfs):
    if ("Episodes" not in dfs) or ("provider_info" not in dfs["metadata"]):
        return {}
    else:
        episodes = dfs["Episodes"]
        provider_info = dfs["metadata"]["provider_info"]

        # merge
        episodes["index_eps"] = episodes.index
        episodes = episodes[episodes["URN"].notna() & (episodes["URN"] != "XXXXXXX")]
        episodes = episodes.merge(
            provider_info, on="URN", how="left", suffixes=["_eps", "_lookup"]
        )
        # If <URN> provided and <URN> not = 'XXXXXXX', then <PLACE_PROVIDER> must = URN Lookup <PLACE_PROVIDER>
        valid = pd.Series(
            [
                (
                    pl_pr in valid.split(",")
                    if (pd.notna(pl_pr) and pd.notna(valid))
                    else False
                )
                for pl_pr, valid in zip(
                    episodes["PLACE_PROVIDER"], episodes["PROVIDER_CODES"]
                )
            ]
        )
        eps_error_locations = episodes.loc[~valid, "index_eps"]
        return {"Episodes": eps_error_locations.unique().tolist()}


def test_validate():
    import pandas as pd

    fake_data_eps = pd.DataFrame(
        [
            {
                "CHILD": "1111",
                "PLACE_PROVIDER": "PR2",
                "URN": 1,
            },  # 0 fail
            {
                "CHILD": "1111",
                "PLACE_PROVIDER": "PR5",
                "URN": pd.NA,
            },  # 1 ignore
            {
                "CHILD": "1111",
                "PLACE_PROVIDER": "PR0",
                "URN": 3,
            },  # 2 fail
            {
                "CHILD": "2222",
                "PLACE_PROVIDER": "PR3",
                "URN": "XXXXXXX",
            },  # 3 ignore
            {
                "CHILD": "3333",
                "PLACE_PROVIDER": "PR0",
                "URN": 2,
            },  # 4 pass
            {
                "CHILD": "3333",
                "PLACE_PROVIDER": "PR3",
                "URN": 2,
            },  # 5 fail
            {
                "CHILD": "4444",
                "PLACE_PROVIDER": "PR3",
                "URN": 1,
            },  # 6 pass
            {
                "CHILD": "5555",
                "PLACE_PROVIDER": "PR3",
                "URN": 4,
            },  # 7 fail - if PROVIDER_CODES is null something is wrong
            {
                "CHILD": "5555",
                "PLACE_PROVIDER": "PR4",
                "URN": 1,
            },  # 8 pass
        ]
    )
    fake_provider_info = pd.DataFrame(
        [
            {
                "URN": 1,
                "PROVIDER_CODES": "PR1,PR3,PR4",
            },  # 0
            {
                "URN": 2,
                "PROVIDER_CODES": "PR0",
            },  # 1
            {
                "URN": 3,
                "PROVIDER_CODES": "PR1",
            },  # 2
            {
                "URN": 4,
                "PROVIDER_CODES": pd.NA,
            },  # 3
        ]
    )
    metadata = {"provider_info": fake_provider_info}

    fake_dfs = {"Episodes": fake_data_eps, "metadata": metadata}

    result = validate(fake_dfs)

    assert result == {"Episodes": [0, 2, 5, 7]}



import pandas as pd

from lac_validator.rule_engine import rule_definition


@rule_definition(
    code="143",
    message="The reason for new episode code is not a valid code.",
    affected_fields=["RNE"],
    tables=["Episodes"],
)
def validate(dfs):
    if "Episodes" not in dfs:
        return {}

    episodes = dfs["Episodes"]
    code_list = ["S", "P", "L", "T", "U", "B"]

    mask = episodes["RNE"].isin(code_list) | episodes["RNE"].isna()

    validation_error_mask = ~mask
    validation_error_locations = episodes.index[validation_error_mask]

    return {"Episodes": validation_error_locations.tolist()}


def test_validate():
    import pandas as pd

    fake_data = pd.DataFrame(
        {
            "RNE": ["S", "p", "SP", "a", "", pd.NA],
        }
    )

    fake_dfs = {"Episodes": fake_data}

    result = validate(fake_dfs)

    assert result == {"Episodes": [1, 2, 3, 4]}



import pandas as pd

from lac_validator.rule_engine import rule_definition


@rule_definition(
    code="114",
    message="Data entry to record the status of former carer(s) of an adopted child is invalid.",
    affected_fields=["FOSTER_CARE"],
    tables=["AD1"],
)
def validate(dfs):
    if "AD1" not in dfs:
        return {}

    adoptions = dfs["AD1"]
    code_list = ["0", "1"]

    mask = (
        adoptions["FOSTER_CARE"].astype(str).isin(code_list)
        | adoptions["FOSTER_CARE"].isna()
    )

    validation_error_mask = ~mask
    validation_error_locations = adoptions.index[validation_error_mask]

    return {"AD1": validation_error_locations.tolist()}


def test_validate():
    import pandas as pd

    fake_data = pd.DataFrame(
        {
            "FOSTER_CARE": [0, 1, "0", "1", 2, "former foster carer", "", pd.NA],
        }
    )

    fake_dfs = {"AD1": fake_data}

    result = validate(fake_dfs)

    assert result == {"AD1": [4, 5, 6]}



import pandas as pd

from lac_validator.rule_engine import rule_definition


@rule_definition(
    code="225",
    message="Reason for placement change must be recorded.",
    affected_fields=["REASON_PLACE_CHANGE"],
    tables=["Episodes"],
)
def validate(dfs):
    if "Episodes" not in dfs:
        return {}
    else:
        epi = dfs["Episodes"]
        epi["DECOM"] = pd.to_datetime(epi["DECOM"], format="%d/%m/%Y", errors="coerce")
        epi.sort_values(["CHILD", "DECOM"], inplace=True)
        epi.reset_index(inplace=True)
        epi.reset_index(inplace=True)
        epi["LAG_INDEX"] = epi["level_0"].shift(1)
        m_epi = epi.merge(
            epi,
            how="inner",
            left_on="level_0",
            right_on="LAG_INDEX",
            suffixes=["", "_NEXT"],
        )
        m_epi = m_epi[m_epi["CHILD"] == m_epi["CHILD_NEXT"]]

        mask_is_X1 = m_epi["REC"] == "X1"
        mask_null_place_chg = m_epi["REASON_PLACE_CHANGE"].isna()
        mask_place_not_T = ~m_epi["PLACE"].isin(["T0", "T1", "T2", "T3", "T4"])
        mask_next_is_PBTU = m_epi["RNE_NEXT"].isin(["P", "B", "T", "U"])
        mask_next_place_not_T = ~m_epi["PLACE_NEXT"].isin(
            ["T0", "T1", "T2", "T3", "T4"]
        )

        error_mask = (
            mask_is_X1
            & mask_null_place_chg
            & mask_place_not_T
            & mask_next_is_PBTU
            & mask_next_place_not_T
        )

        error_list = m_epi["index"][error_mask].to_list()
        return {"Episodes": error_list}


def test_validate():
    import pandas as pd

    fake_data_epi = pd.DataFrame(
        [
            {
                "CHILD": "111",
                "DECOM": "01/06/2020",
                "RNE": "P",
                "REC": "X1",
                "PLACE": "U1",
                "REASON_PLACE_CHANGE": pd.NA,
            },  # 0 Fails
            {
                "CHILD": "111",
                "DECOM": "05/06/2020",
                "RNE": "P",
                "REC": "E11",
                "PLACE": "X1",
                "REASON_PLACE_CHANGE": pd.NA,
            },  # 1
            {
                "CHILD": "111",
                "DECOM": "06/06/2020",
                "RNE": "B",
                "REC": "E3",
                "PLACE": "X1",
                "REASON_PLACE_CHANGE": pd.NA,
            },  # 2
            {
                "CHILD": "123",
                "DECOM": "08/06/2020",
                "RNE": "P",
                "REC": "X1",
                "PLACE": "U1",
                "REASON_PLACE_CHANGE": pd.NA,
            },  # 3 Fails
            {
                "CHILD": "123",
                "DECOM": "10/06/2020",
                "RNE": "P",
                "REC": "E3",
                "PLACE": "X1",
                "REASON_PLACE_CHANGE": pd.NA,
            },  # 4
            {
                "CHILD": "333",
                "DECOM": "06/06/2020",
                "RNE": "T",
                "REC": "X1",
                "PLACE": "U1",
                "REASON_PLACE_CHANGE": pd.NA,
            },  # 5
            {
                "CHILD": "333",
                "DECOM": "07/06/2020",
                "RNE": "L",
                "REC": "X1",
                "PLACE": "X1",
                "REASON_PLACE_CHANGE": "CHANGE",
            },  # 6 Passes as RPC not null
            {
                "CHILD": "333",
                "DECOM": "08/06/2020",
                "RNE": "P",
                "REC": "E3",
                "PLACE": "U1",
                "REASON_PLACE_CHANGE": pd.NA,
            },  # 7
            {
                "CHILD": "444",
                "DECOM": "09/06/2020",
                "RNE": "U",
                "REC": "X1",
                "PLACE": "T1",
                "REASON_PLACE_CHANGE": pd.NA,
            },  # 8
            {
                "CHILD": "444",
                "DECOM": "15/06/2020",
                "RNE": "P",
                "REC": "X1",
                "PLACE": "X1",
                "REASON_PLACE_CHANGE": pd.NA,
            },  # 9 Passes next place T3
            {
                "CHILD": "444",
                "DECOM": "16/06/2020",
                "RNE": "P",
                "REC": "E3",
                "PLACE": "T3",
                "REASON_PLACE_CHANGE": pd.NA,
            },  # 10
            {
                "CHILD": "666",
                "DECOM": "17/06/2020",
                "RNE": "P",
                "REC": pd.NA,
                "PLACE": "T4",
                "REASON_PLACE_CHANGE": pd.NA,
            },  # 11
        ]
    )

    fake_dfs = {"Episodes": fake_data_epi}

    result = validate(fake_dfs)

    assert result == {"Episodes": [0, 3]}



import pandas as pd

from lac_validator.rule_engine import rule_definition


@rule_definition(
    code="570",
    message="The date that the child started to be missing or away from placement without authorisation is after the end of the collection year.",
    affected_fields=["MIS_START"],
    tables=["Missing"],
)
def validate(dfs):
    if "Missing" not in dfs:
        return {}
    else:
        mis = dfs["Missing"]
        collection_end = pd.to_datetime(
            dfs["metadata"]["collection_end"], format="%d/%m/%Y", errors="coerce"
        )

        mis["MIS_START"] = pd.to_datetime(
            mis["MIS_START"], format="%d/%m/%Y", errors="coerce"
        )
        error_mask = mis["MIS_START"] > collection_end

        return {"Missing": mis.index[error_mask].to_list()}


def test_validate():
    import pandas as pd

    fake_data = pd.DataFrame(
        {
            "MIS_START": [
                "08/04/2020",
                "22/06/2020",
                pd.NA,
                "13/10/2005",
                "10/05/2001",
            ],
        }
    )

    metadata = {"collection_end": "31/03/2020"}

    fake_dfs = {"Missing": fake_data, "metadata": metadata}

    result = validate(fake_dfs)

    assert result == {"Missing": [0, 1]}



import pandas as pd

from lac_validator.rule_engine import rule_definition


@rule_definition(
    code="391",
    message="Young person was not 17, 18, 19, 20 or 21 during the current collection year. ",
    affected_fields=["DOB", "IN_TOUCH", "ACTIV", "ACCOM"],
    tables=["OC3"],
)
def validate(dfs):
    if "OC3" not in dfs:
        return {}
    else:
        oc3 = dfs["OC3"]
        collection_end = dfs["metadata"]["collection_end"]

        # convert dates to datetime format
        oc3["DOB"] = pd.to_datetime(oc3["DOB"], format="%d/%m/%Y", errors="coerce")
        collection_end = pd.to_datetime(
            collection_end, format="%d/%m/%Y", errors="coerce"
        )

        # If <DOB> < 17 years prior to <COLLECTION_END_DATE> then <IN_TOUCH>, <ACTIV> and <ACCOM> should not be provided
        check_age = oc3["DOB"] + pd.offsets.DateOffset(years=17) > collection_end
        mask = check_age & (
            oc3["IN_TOUCH"].notna() | oc3["ACTIV"].notna() | oc3["ACCOM"].notna()
        )
        # Then raise an error if either IN_TOUCH, ACTIV, or ACCOM have been provided too

        # error locations
        oc3_error_locs = oc3.index[mask]

        return {"OC3": oc3_error_locs.tolist()}


def test_validate():
    import pandas as pd

    fake_data_oc3 = pd.DataFrame(
        {
            "CHILD": ["A", "B", "C", "D", "E"],
            "DOB": [
                "01/01/2001",
                "01/01/2016",
                "20/12/1997",
                "01/01/2002",
                "03/01/2004",
            ],
            "IN_TOUCH": ["DIED", "Yes", "RHOM", pd.NA, pd.NA],
            "ACTIV": [pd.NA, pd.NA, "XXX", pd.NA, pd.NA],
            "ACCOM": [pd.NA, pd.NA, pd.NA, "XXX", pd.NA],
        }
    )
    metadata = {"collection_end": "31/03/2018"}

    fake_dfs = {"OC3": fake_data_oc3, "metadata": metadata}

    result = validate(fake_dfs)
    assert result == {"OC3": [1, 3]}



import pandas as pd

from lac_validator.rule_engine import rule_definition


@rule_definition(
    code="SW05STG1",
    message="Date social worker episode began is not a valid date.",
    affected_fields=["SW_DEC"],
    tables=["SWEpisodes"],
)
def validate(dfs):
    if "SWEpisodes" not in dfs:
        return {}
    else:
        import lac_validator.rules.rule_utils

        df = dfs["SWEpisodes"]

        df = df[df["SW_DEC"].notna()].copy()

        df["SW_DEC_dt"] = df["SW_DEC"].apply(lac_validator.rules.rule_utils.valid_date)

        error_rows = df[df["SW_DEC_dt"].isna()].index

        return {"SWEpisodes": error_rows.tolist()}


def test_validate():
    import pandas as pd

    fake_data = pd.DataFrame(
        {
            "SW_DEC": ["ZZ/ZZ/ZZZZ", "01/01/2001", "zz", "01/01/ZZZZ", pd.NA],
        }
    )

    fake_dfs = {"SWEpisodes": fake_data}

    result = validate(fake_dfs)

    assert result == {"SWEpisodes": [2]}



import pandas as pd

from lac_validator.rule_engine import rule_definition


@rule_definition(
    code="620",
    message="Child has been recorded as a mother, but date of birth shows that the mother is under 11 years of age.",
    affected_fields=["DOB", "MOTHER"],
    tables=["Header"],
)
def validate(dfs):
    if "Header" not in dfs:
        return {}
    else:
        hea = dfs["Header"]
        collection_start = pd.to_datetime(
            dfs["metadata"]["collection_start"], format="%d/%m/%Y", errors="coerce"
        )
        hea["DOB"] = pd.to_datetime(hea["DOB"], format="%d/%m/%Y", errors="coerce")

        hea_mother = hea[hea["MOTHER"].astype(str) == "1"]
        error_cohort = (
            hea_mother["DOB"] + pd.offsets.DateOffset(years=11)
        ) > collection_start

        return {"Header": hea_mother.index[error_cohort].to_list()}


def test_validate():
    import pandas as pd

    fake_hea = pd.DataFrame(
        [
            {"MOTHER": "1", "DOB": pd.NA},  # 0
            {"MOTHER": "1", "DOB": "07/02/2020"},  # 1  Fails
            {"MOTHER": "1", "DOB": "03/02/2020"},  # 2  Fails
            {"MOTHER": pd.NA, "DOB": pd.NA},  # 3
            {"MOTHER": "1", "DOB": "18/01/1981"},  # 4  Passes old DOB
            {"MOTHER": "0", "DOB": "13/02/2020"},  # 5
        ]
    )

    metadata = {"collection_start": "01/04/2020"}

    fake_dfs = {"Header": fake_hea, "metadata": metadata}

    result = validate(fake_dfs)

    assert result == {"Header": [1, 2]}



import pandas as pd

from lac_validator.rule_engine import rule_definition


@rule_definition(
    code="113",
    message="Date matching child and adopter(s) is not a valid date.",
    affected_fields=["DATE_MATCH"],
    tables=["AD1"],
)
def validate(dfs):
    if "AD1" not in dfs:
        return {}
    else:
        ad1 = dfs["AD1"]
        mask = pd.to_datetime(
            ad1["DATE_MATCH"], format="%d/%m/%Y", errors="coerce"
        ).notna()

        na_location = ad1["DATE_MATCH"].isna()

        validation_error_mask = ~mask & ~na_location
        validation_error_locations = ad1.index[validation_error_mask]

        return {"AD1": validation_error_locations.tolist()}


def test_validate():
    import pandas as pd

    fake_data = pd.DataFrame(
        {
            "DATE_MATCH": ["22/11/2015", "08/05/2010", "38/04/2019", "01/01/19", pd.NA],
        }
    )

    fake_dfs = {"AD1": fake_data}

    result = validate(fake_dfs)

    assert result == {"AD1": [2, 3]}



import pandas as pd

from lac_validator.rule_engine import rule_definition


@rule_definition(
    code="583",
    message="More than one date the child should be placed for adoption has been reported, but an earlier decision has not been revoked. Check that both dates are required and add in the date the earlier decision was revoked. Note this may be in a previous year.",
    affected_fields=["DATE_PLACED", "DATE_PLACED_CEASED"],
    tables=["PlacedAdoption"],
)
def validate(dfs):
    # If more than one <DATE_PLACED> is present, then all but the latest <DATE_PLACED> must have <DATE_PLACED_CEASED>
    # (The latest <DATE_PLACED> may or may not have <DATE_PLACED_CEASED>)
    # Apply for 2023 onwards
    if "PlacedAdoption" not in dfs:
        return {}
    else:
        # # If a child has more than one Na in date_placed_cease, they fail, sending all Na rows as errors
        df = dfs["PlacedAdoption"]

        df_ordered = df.sort_values("DATE_PLACED", ascending=False)
        df_ordered.drop_duplicates("CHILD", keep="first", inplace=True)
        df_ordered_index = list(df_ordered.index)
        df_dropped_recent = df[
            (~df.index.isin(df_ordered_index)) & df["DATE_PLACED_CEASED"].isna()
        ]
        non_recent_nans = list(df_dropped_recent.index)
        error_mask = df.index.isin(non_recent_nans)

        return {"PlacedAdoption": df.index[error_mask].to_list()}


def test_validate():
    import pandas as pd

    fake_data = pd.DataFrame(
        [
            {
                "CHILD": "child1",
                "DATE_PLACED": "01/02/2020",
                "DATE_PLACED_CEASED": "02/02/2020",
            },  # 0 pass
            {
                "CHILD": "child1",
                "DATE_PLACED": "03/02/2020",
                "DATE_PLACED_CEASED": "04/02/2020",
            },  # 1 pass
            {
                "CHILD": "child2",
                "DATE_PLACED": "01/02/2020",
                "DATE_PLACED_CEASED": pd.NA,
            },  # 2 fail
            {
                "CHILD": "child2",
                "DATE_PLACED": "02/02/2020",
                "DATE_PLACED_CEASED": pd.NA,
            },  # 3 Pass
            {
                "CHILD": "child3",
                "DATE_PLACED": "01/02/2020",
                "DATE_PLACED_CEASED": "02/02/2020",
            },  # 4 pass
            {
                "CHILD": "child3",
                "DATE_PLACED": "03/02/2020",
                "DATE_PLACED_CEASED": pd.NA,
            },  # 5 pass
            {
                "CHILD": "child4",
                "DATE_PLACED": "01/02/2020",
                "DATE_PLACED_CEASED": "02/02/2020",
            },  # 6 pass
            {
                "CHILD": "child4",
                "DATE_PLACED": "01/01/2020",
                "DATE_PLACED_CEASED": pd.NA,
            },  # 7 fail
        ]
    )

    fake_data["DATE_PLACED"] = pd.to_datetime(
        fake_data["DATE_PLACED"], format="%d/%m/%Y"
    )
    fake_data["DATE_PLACED_CEASED"] = pd.to_datetime(
        fake_data["DATE_PLACED_CEASED"], format="%d/%m/%Y"
    )

    fake_dfs = {"PlacedAdoption": fake_data}

    assert validate(fake_dfs) == {"PlacedAdoption": [2, 7]}



import pandas as pd

from lac_validator.rule_engine import rule_definition


@rule_definition(
    code="1003",
    message="Date of LA's decision that a child should be placed for adoption is before the child started to be looked after.",
    affected_fields=["DATE_PLACED", "DECOM", "RNE"],
    tables=["Episodes", "PlacedAdoption"],
)
def validate(dfs):
    if "Episodes" not in dfs or "PlacedAdoption" not in dfs:
        return {}
    else:
        episodes = dfs["Episodes"]
        placed_adoption = dfs["PlacedAdoption"]

        # to datetime
        placed_adoption["DATE_PLACED"] = pd.to_datetime(
            placed_adoption["DATE_PLACED"], format="%d/%m/%Y", errors="coerce"
        )
        episodes["DECOM"] = pd.to_datetime(
            episodes["DECOM"], format="%d/%m/%Y", errors="coerce"
        )

        # select the earliest episodes with RNE =  S
        eps_rne = episodes[episodes["RNE"] == "S"]
        first_eps_idxs = eps_rne.groupby("CHILD")["DECOM"].idxmin()
        first_eps = eps_rne.loc[first_eps_idxs]
        # prepare to merge
        placed_adoption.reset_index(inplace=True)
        first_eps.reset_index(inplace=True)
        merged = first_eps.merge(
            placed_adoption, how="left", on="CHILD", suffixes=["_eps", "_pa"]
        )

        # <DATE_PLACED> cannot be prior to <DECOM> of the first episode with <RNE> = 'S'
        mask = merged["DATE_PLACED"] < merged["DECOM"]
        eps_error_locs = merged.loc[mask, "index_eps"]
        pa_error_locs = merged.loc[mask, "index_pa"]
        return {
            "Episodes": eps_error_locs.tolist(),
            "PlacedAdoption": pa_error_locs.tolist(),
        }


def test_validate():
    import pandas as pd

    fake_data_episodes = pd.DataFrame(
        [
            {"CHILD": 101, "DECOM": "01/03/1980", "RNE": "S"},  # 0 fail
            {"CHILD": 102, "DECOM": "01/03/1980", "RNE": "o"},
            {
                "CHILD": 102,
                "DECOM": "01/03/1980",
                "RNE": "S",
            },  # 2 ignore DATE_PLACED is nan
            {"CHILD": 103, "DECOM": "01/02/1970", "RNE": "o"},
            {"CHILD": 103, "DECOM": "01/03/1979", "RNE": "S"},  # 4 fail
            {"CHILD": 103, "DECOM": "01/01/1981", "RNE": "S"},
            {"CHILD": 104, "DECOM": "01/03/1979", "RNE": "o"},  # ignore no RNE is S
            {"CHILD": 104, "DECOM": "01/01/1981", "RNE": "o"},
            {"CHILD": 105, "DECOM": "01/03/1979", "RNE": "o"},
            {"CHILD": 105, "DECOM": "01/01/1981", "RNE": "o"},
            {"CHILD": 105, "DECOM": "01/01/1981", "RNE": "S"},  # 10 pass
        ]
    )
    fake_placed_adoption = pd.DataFrame(
        [
            {"CHILD": 101, "DATE_PLACED": "26/05/1978"},  # 0 fail
            {"CHILD": 102, "DATE_PLACED": pd.NA},  # 1
            {"CHILD": 103, "DATE_PLACED": "26/05/1970"},  # 2 fail
            {"CHILD": 104, "DATE_PLACED": "01/02/1960"},  # 3
            {"CHILD": 105, "DATE_PLACED": "26/05/2019"},  # 4
        ]
    )
    fake_dfs = {"Episodes": fake_data_episodes, "PlacedAdoption": fake_placed_adoption}

    result = validate(fake_dfs)
    assert result == {"Episodes": [0, 4], "PlacedAdoption": [0, 2]}



import pandas as pd

from lac_validator.rule_engine import rule_definition


@rule_definition(
    code="SW01aSTG1",
    message="For each social worker episode, information should be complete",
    affected_fields=["SW_DECOM", "SW_ID"],
    tables=["SWEpisodes"],
)
def validate(dfs):
    if "SWEpisodes" not in dfs:
        return {}
    else:
        df = dfs["SWEpisodes"]

        # If <SW_DECOM> < 01 April 2023 then at least one instance of each of the following items should be provided:
        # <SW_ID>, <SW_DECOM>
        #  Note: <SW_REASON> may or may not have been provided

        df["SW_DECOM"] = pd.to_datetime(
            df["SW_DECOM"], format="%d/%m/%Y", errors="coerce"
        )

        no_items = df["SW_ID"].isna()
        before_1_apr_23 = df["SW_DECOM"] < pd.to_datetime("01/04/2023", dayfirst=True)
        no_decom = df["SW_DECOM"].isna()

        errors = df[(before_1_apr_23 & no_items) | no_decom]

        error_rows = errors.index.tolist()

        return {"SWEpisodes": error_rows}


def test_validate():
    import pandas as pd

    fake_swe = pd.DataFrame(
        [
            {
                "CHILD": "child1",
                "SW_ID": pd.NA,
                "SW_DECOM": "01/01/2024",
                "xx": "xx",
            },  # ignore after 1 apr 2023
            {
                "CHILD": "child2",
                "SW_ID": "xx",
                "SW_DECOM": pd.NA,
                "xx": "xx",
            },  # fail no decom
            {
                "CHILD": "child3",
                "SW_ID": "xx",
                "SW_DECOM": "01/01/2024",
                "xx": pd.NA,
            },  # ignore after 1 apr 2023
            {
                "CHILD": "child4",
                "SW_ID": "xx",
                "SW_DECOM": "01/01/1999",
                "xx": "xx",
            },  # ignore before apr 1 2023 but no reason
            {
                "CHILD": "child5",
                "SW_ID": "xx",
                "SW_DECOM": "01/01/1999",
                "xx": pd.NA,
            },  # ignore before apr 1 2023 but no reason
            {
                "CHILD": "child7",
                "SW_ID": "xx",
                "SW_DECOM": "01/01/2024",
                "SW_REASON": pd.NA,
            },  # ignore after 1 apr 2023
            {
                "CHILD": "child9",
                "SW_ID": "xx",
                "xx": "xx",
                "SW_REASON": pd.NA,
            },  # fail, no details or decom
        ]
    )

    fake_dfs = {"SWEpisodes": fake_swe}

    result = validate(fake_dfs)

    assert result == {"SWEpisodes": [1, 6]}



import pandas as pd

from lac_validator.rule_engine import rule_definition


@rule_definition(
    code="227",
    message="Ofsted Unique reference number (URN) is not valid for the episode start date.",
    affected_fields=["URN", "DECOM"],
    tables=["Episodes", "Provider Info"],
)
def validate(dfs):
    if ("Episodes" not in dfs) or ("provider_info" not in dfs["metadata"]):
        return {}
    else:
        episodes = dfs["Episodes"]
        provider_info = dfs["metadata"]["provider_info"]

        # convert date fields from strings to datetime format. NB. REG_END is in datetime format already.
        episodes["DECOM"] = pd.to_datetime(
            episodes["DECOM"], format="%d/%m/%Y", errors="coerce"
        )

        # merge
        episodes["index_eps"] = episodes.index
        episodes = episodes[episodes["URN"].notna() & (episodes["URN"] != "XXXXXXX")]
        merged = episodes.merge(provider_info, on="URN", how="left")
        # If <URN> provided and <URN> not = 'XXXXXXX', then if <URN> and <REG_END> are provided then <DECOM> must be before <REG_END>
        mask = merged["REG_END"].notna() & (merged["DECOM"] >= merged["REG_END"])

        eps_error_locations = merged.loc[mask, "index_eps"]
        return {"Episodes": eps_error_locations.tolist()}


def test_validate():
    import pandas as pd

    fake_data_eps = pd.DataFrame(
        [
            {
                "CHILD": "1111",
                "DECOM": "01/01/2014",
                "URN": 1,
            },  # 0 pass
            {
                "CHILD": "1111",
                "DECOM": "01/02/2015",
                "URN": pd.NA,
            },  # 1 ignore: URN not provided
            {
                "CHILD": "1111",
                "DECOM": "01/01/2016",
                "URN": 3,
            },  # 2 pass
            {
                "CHILD": "2222",
                "DECOM": "01/01/2010",
                "URN": "XXXXXXX",
            },  # 3 ignore
            {
                "CHILD": "3333",
                "DECOM": "01/01/2010",
                "URN": 2,
            },  # 4 pass
            {
                "CHILD": "3333",
                "DECOM": "25/12/2015",
                "URN": 2,
            },  # 5 fail DECOM after REG_END
            {
                "CHILD": "4444",
                "DECOM": "25/12/2016",
                "URN": 1,
            },  # 6 fail. DECOM after REG_END
            {
                "CHILD": "5555",
                "DECOM": "01/01/2010",
                "URN": 4,
            },  # 7 ignore: REG_END not provided
            {
                "CHILD": "5555",
                "DECOM": "25/12/2015",
                "URN": 1,
            },  # 8 fail DECOM equals REG_END
        ]
    )
    provider_info = pd.DataFrame(
        [
            {
                "URN": 1,
                "REG_END": "25/12/2015",
            },  # 0
            {
                "URN": 2,
                "REG_END": "21/02/2014",
            },  # 1
            {
                "URN": 3,
                "REG_END": "25/12/2017",
            },  # 2
            {
                "URN": 3,
                "REG_END": pd.NA,
            },  # 3
        ]
    )
    provider_info["REG_END"] = pd.to_datetime(
        provider_info["REG_END"], format="%d/%m/%Y", errors="coerce"
    )

    metadata = {"provider_info": provider_info}

    fake_dfs = {"Episodes": fake_data_eps, "metadata": metadata}

    result = validate(fake_dfs)

    assert result == {"Episodes": [5, 6, 8]}



import pandas as pd

from lac_validator.rule_engine import rule_definition


@rule_definition(
    code="146",
    message="Placement type code is not a valid code.",
    affected_fields=["PLACE"],
    tables=["Episodes"],
)
def validate(dfs):
    if "Episodes" not in dfs:
        return {}

    episodes = dfs["Episodes"]
    code_list = [
        "A3",
        "A4",
        "A5",
        "A6",
        "H5",
        "K1",
        "K2",
        "P1",
        "P2",
        "P3",
        "R1",
        "R2",
        "R3",
        "R5",
        "S1",
        "T0",
        "T1",
        "T2",
        "T3",
        "T4",
        "U1",
        "U2",
        "U3",
        "U4",
        "U5",
        "U6",
        "Z1",
    ]

    mask = episodes["PLACE"].isin(code_list) | episodes["PLACE"].isna()

    validation_error_mask = ~mask
    validation_error_locations = episodes.index[validation_error_mask]

    return {"Episodes": validation_error_locations.tolist()}


def test_validate():
    import pandas as pd

    fake_data = pd.DataFrame(
        {
            "PLACE": ["A2", "R4", "Z", "P1", "", "t3", pd.NA],
        }
    )

    fake_dfs = {"Episodes": fake_data}

    result = validate(fake_dfs)

    assert result == {"Episodes": [0, 1, 2, 4, 5]}



import pandas as pd

from lac_validator.rule_engine import rule_definition


@rule_definition(
    code="207",
    message="Mother status for the current year disagrees with the mother status already recorded for this child.",
    affected_fields=["MOTHER"],
    tables=["Episodes", "Header", "Header_last"],
)
def validate(dfs):
    if "Header" not in dfs or "Header_last" not in dfs or "Episodes" not in dfs:
        return {}
    else:
        header = dfs["Header"]
        header_last = dfs["Header_last"]
        episodes = dfs["Episodes"]

        header.reset_index(inplace=True)

        header_merged = header.merge(
            header_last,
            how="left",
            on=["CHILD"],
            suffixes=("", "_last"),
            indicator=True,
        )

        header_merged = header_merged.merge(
            episodes[["CHILD"]],
            on="CHILD",
            how="left",
            suffixes=("", "_eps"),
            indicator="_eps",
        )

        in_both_years = header_merged["_merge"] == "both"
        has_no_episodes = header_merged["_eps"] == "left_only"
        mother_is_different = header_merged["MOTHER"].astype(str) != header_merged[
            "MOTHER_last"
        ].astype(str)
        mother_was_true = header_merged["MOTHER_last"].astype(str) == "1"

        error_mask = (
            in_both_years & ~has_no_episodes & mother_is_different & mother_was_true
        )

        error_locations = list(header_merged.loc[error_mask, "index"].unique())

        return {"Header": error_locations}


def test_validate():
    import pandas as pd

    fake_data = pd.DataFrame(
        {
            "CHILD": [
                "101",
                "102",
                "103",
                "104",
                "105",
                "106",
                "108",
                "109",
                "110",
                "111",
            ],
            "MOTHER": ["1", 0, "1", pd.NA, pd.NA, "1", pd.NA, "0", "2", pd.NA],
        }
    )

    fake_data_prev = pd.DataFrame(
        {
            "CHILD": [
                "101",
                "102",
                "103",
                "104",
                "105",
                "107",
                "108",
                "109",
                "110",
                "111",
            ],
            "MOTHER": ["1", 1, "0", 1, pd.NA, "1", "0", pd.NA, "1", "1"],
        }
    )

    fake_episodes = pd.DataFrame(
        {
            "CHILD": [
                "101",
                "101",
                "102",
                "103",
                "104",
                "105",
                "106",
                "107",
                "108",
                "109",
                "110",
            ],
        }
    )

    fake_dfs = {
        "Header": fake_data,
        "Header_last": fake_data_prev,
        "Episodes": fake_episodes,
    }

    result = validate(fake_dfs)

    assert result == {"Header": [1, 3, 8]}



import pandas as pd

from lac_validator.rule_engine import rule_definition


@rule_definition(
    code="189",
    message="Child is aged 17 years or over at the beginning of the year, but an Strengths and Difficulties "
    + "(SDQ) score or a reason for no Strengths and Difficulties (SDQ) score has been completed.",
    affected_fields=["DOB", "SDQ_SCORE", "SDQ_REASON"],
    tables=["OC2"],
)
def validate(dfs):
    if "OC2" not in dfs:
        return {}
    else:
        oc2 = dfs["OC2"]
        collection_start = dfs["metadata"]["collection_start"]

        # datetime format allows appropriate comparison between dates
        oc2["DOB"] = pd.to_datetime(oc2["DOB"], format="%d/%m/%Y", errors="coerce")
        collection_start = pd.to_datetime(
            collection_start, format="%d/%m/%Y", errors="coerce"
        )

        # If <DOB> >17 years prior to <COLLECTION_START_DATE> then <SDQ_SCORE> and <SDQ_REASON> should not be provided
        mask = ((oc2["DOB"] + pd.offsets.DateOffset(years=17)) < collection_start) & (
            oc2["SDQ_REASON"].notna() | oc2["SDQ_SCORE"].notna()
        )
        # That is, raise error if collection_start > DOB + 17years
        oc_error_locs = oc2.index[mask]
        return {"OC2": oc_error_locs.tolist()}


def test_validate():
    import pandas as pd

    metadata = {"collection_start": "01/04/1980"}

    oc2 = pd.DataFrame(
        {
            "CHILD": ["0", "1", "2", "3"],
            "DOB": ["01/04/1962", "02/04/1963", "01/01/1970", "31/03/1960"],
            "SDQ_SCORE": ["!!!", "OK", pd.NA, pd.NA],
            "SDQ_REASON": [pd.NA, pd.NA, pd.NA, pd.NA],
        }
    )

    fake_dfs = {"OC2": oc2, "metadata": metadata}

    result = validate(fake_dfs)

    assert result == {
        "OC2": [
            0,
        ]
    }



import pandas as pd

from lac_validator.rule_engine import rule_definition


@rule_definition(
    code="561",
    message="Date of the decision that the child should be placed for adoption this year is the same as that recorded last year but records show that the decision changed, and the child should no longer be placed for adoption last year.",
    affected_fields=["DATE_PLACED"],
    tables=["PlacedAdoption", "PlacedAdoption_last"],
)
def validate(dfs):
    if "PlacedAdoption" not in dfs or "PlacedAdoption_last" not in dfs:
        return {}
    else:
        placed_adoption = dfs["PlacedAdoption"]
        pa_last = dfs["PlacedAdoption_last"]

        # prepare to merge
        placed_adoption.reset_index(inplace=True)
        pa_last.reset_index(inplace=True)
        merged = placed_adoption.merge(
            pa_last, how="inner", on="CHILD", suffixes=["_now", "_last"]
        )

        # If <CURRENT_COLLECTION_YEAR> <DATE_PLACED> is = <CURRENT_COLLECTION_YEAR> -1 <DATE_PLACED> then <CURRENT_COLLECTION_YEAR> -1 <DATE_PLACED_CEASED> and <REASON_PLACED_CEASED> should be Null
        mask = (merged["DATE_PLACED_now"] == merged["DATE_PLACED_last"]) & merged[
            ["REASON_PLACED_CEASED_last", "DATE_PLACED_CEASED_last"]
        ].notna().any(axis=1)

        # error locations
        error_locs = merged.loc[mask, "index_now"]
        return {"PlacedAdoption": error_locs.tolist()}


def test_validate():
    import pandas as pd

    fake_placed_adoption = pd.DataFrame(
        [
            {
                "CHILD": 101,
                "DATE_PLACED_CEASED": "26/05/2000",
                "DATE_PLACED": "26/05/2019",
                "REASON_PLACED_CEASED": "xx",
            },
            # 0
            {
                "CHILD": 102,
                "DATE_PLACED_CEASED": "01/07/2018",
                "DATE_PLACED": "26/05/2000",
                "REASON_PLACED_CEASED": "xx",
            },
            # 1
            {
                "CHILD": 103,
                "DATE_PLACED_CEASED": "26/05/2000",
                "DATE_PLACED": pd.NA,
                "REASON_PLACED_CEASED": "xx",
            },  # 2
            {
                "CHILD": 104,
                "DATE_PLACED_CEASED": "26/05/2017",
                "DATE_PLACED": "01/02/2016",
                "REASON_PLACED_CEASED": "xx",
            },
            # 3
            {
                "CHILD": 106,
                "DATE_PLACED_CEASED": pd.NA,
                "DATE_PLACED": "26/05/2029",
                "REASON_PLACED_CEASED": "xx",
            },  # 4
        ]
    )

    fake_pa_last = pd.DataFrame(
        [
            {
                "CHILD": 101,
                "DATE_PLACED_CEASED": pd.NA,
                "DATE_PLACED": "26/05/2019",
                "REASON_PLACED_CEASED": "xx",
            },
            # 0 fail
            {
                "CHILD": 102,
                "DATE_PLACED_CEASED": "01/07/2018",
                "DATE_PLACED": "26/05/2000",
                "REASON_PLACED_CEASED": "xx",
            },
            # 1 fail
            {
                "CHILD": 103,
                "DATE_PLACED_CEASED": "26/05/2000",
                "DATE_PLACED": pd.NA,
                "REASON_PLACED_CEASED": "xx",
            },  # 2
            {
                "CHILD": 104,
                "DATE_PLACED_CEASED": "26/05/2017",
                "DATE_PLACED": "01/02/2016",
                "REASON_PLACED_CEASED": "xx",
            },
            # 3
            {
                "CHILD": 105,
                "DATE_PLACED_CEASED": pd.NA,
                "DATE_PLACED": "26/05/2019",
                "REASON_PLACED_CEASED": pd.NA,
            },
            # 4 pass
            {
                "CHILD": 106,
                "DATE_PLACED_CEASED": "xx",
                "DATE_PLACED": "01/07/2019",
                "REASON_PLACED_CEASED": "xx",
            },
            # 5 pass
        ]
    )

    fake_dfs = {
        "PlacedAdoption": fake_placed_adoption,
        "PlacedAdoption_last": fake_pa_last,
    }

    result = validate(fake_dfs)
    assert result == {"PlacedAdoption": [0, 1, 3]}



import pandas as pd

from lac_validator.rule_engine import rule_definition


@rule_definition(
    code="302",
    message="First episode starts before child was born.",
    affected_fields=["DECOM", "DOB"],
    tables=["Header", "Episodes"],
)
def validate(dfs):
    if "Header" not in dfs:
        return {}
    if "Episodes" not in dfs:
        return {}
    else:
        header = dfs["Header"]
        episodes = dfs["Episodes"]

        header["DOB"] = pd.to_datetime(
            header["DOB"], format="%d/%m/%Y", errors="coerce"
        )
        episodes["DECOM"] = pd.to_datetime(
            episodes["DECOM"], format="%d/%m/%Y", errors="coerce"
        )

        episodes = episodes.reset_index()
        header = header.reset_index()

        episodes = episodes.loc[episodes.groupby("CHILD")["DECOM"].idxmin()]

        merged = episodes.merge(
            header, how="left", on=["CHILD"], suffixes=("_eps", "_hdr")
        )

        # omitting looking for the 'S' episode as we may not have it in current year's data
        # care_start = merged['RNE'].str.upper().astype(str).isin(['S'])

        started_before_born = merged["DOB"] > merged["DECOM"]

        eps_errors = merged.loc[started_before_born, "index_eps"].to_list()
        hdr_errors = merged.loc[started_before_born, "index_hdr"].to_list()
        return {"Episodes": eps_errors, "Header": hdr_errors}


def test_validate():
    import pandas as pd

    fake_data = pd.DataFrame(
        {
            "CHILD": ["101", "102", "101", "102", "103", "104", "104"],
            "RNE": ["S", "S", "X1", pd.NA, "S", "X", "X"],
            "DECOM": [
                "16/03/2021",
                "17/06/2020",
                "20/03/2020",
                pd.NA,
                "23/08/2020",
                "01/02/1988",
                "03/04/1994",
            ],
        }
    )

    fake_data_child = pd.DataFrame(
        {
            "CHILD": ["101", "102", "103", "104"],
            "DOB": ["16/03/2021", "23/09/2019", "31/12/2020", "01/01/1995"],
        }
    )

    fake_dfs = {"Episodes": fake_data, "Header": fake_data_child}

    result = validate(fake_dfs)

    assert result == {"Episodes": [2, 4, 5], "Header": [0, 2, 3]}



import pandas as pd

from lac_validator.rule_engine import rule_definition
from lac_validator.utils import (
    add_col_to_tables_CONTINUOUSLY_LOOKED_AFTER as add_CLA_column,  # Check 'Episodes' present before use!
)


@rule_definition(
    code="186",
    message="Children aged 4 or over at the start of the year and children aged under 17 at the "
    + "end of the year and who have been looked after for at least 12 months continuously "
    + "should have a Strengths and Difficulties (SDQ) score completed.",
    affected_fields=["SDQ_SCORE"],
    tables=["Episodes", "OC2"],
)
def validate(dfs):
    if "Episodes" not in dfs or "OC2" not in dfs:
        return {}

    oc2 = dfs["OC2"]

    collection_start_str = dfs["metadata"]["collection_start"]
    collection_end_str = dfs["metadata"]["collection_end"]

    collection_start = pd.to_datetime(
        collection_start_str, format="%d/%m/%Y", errors="coerce"
    )
    collection_end = pd.to_datetime(
        collection_end_str, format="%d/%m/%Y", errors="coerce"
    )
    oc2["DOB_dt"] = pd.to_datetime(oc2["DOB"], format="%d/%m/%Y", errors="coerce")

    oc2 = add_CLA_column(dfs, "OC2")

    oc2["4th_bday"] = oc2["DOB_dt"] + pd.DateOffset(years=4)
    oc2["17th_bday"] = oc2["DOB_dt"] + pd.DateOffset(years=17)
    error_mask = (
        (oc2["4th_bday"] <= collection_start)
        & (oc2["17th_bday"] > collection_end)
        & oc2["CONTINUOUSLY_LOOKED_AFTER"]
        & oc2[["SDQ_SCORE", "SDQ_REASON"]].isna().all(axis=1)
    )

    oc2_errors = oc2.loc[error_mask].index.to_list()

    return {"OC2": oc2_errors}


def test_validate():
    import pandas as pd

    metadata = {"collection_start": "01/04/1980", "collection_end": "31/03/1981"}

    eps = pd.DataFrame(
        [
            {
                "CHILD": "1",
                "DECOM": "31/03/1980",
                "DEC": "31/12/1980",
                "LS": "C2",
                "REC": "X1",
                "RNE": "o",
            },  # Pass
            {
                "CHILD": "1",
                "DECOM": "31/12/1980",
                "DEC": pd.NA,
                "LS": "C2",
                "REC": pd.NA,
                "RNE": "o",
            },  # Pass
            {
                "CHILD": "2",
                "DECOM": "31/03/1980",
                "DEC": pd.NA,
                "LS": "C2",
                "REC": pd.NA,
                "RNE": "o",
            },  # Pass
            {
                "CHILD": "3",
                "DECOM": "31/03/1980",
                "DEC": pd.NA,
                "LS": "C2",
                "REC": pd.NA,
                "RNE": "o",
            },  # Pass
            {
                "CHILD": "4",
                "DECOM": "31/03/1980",
                "DEC": pd.NA,
                "LS": "C2",
                "REC": pd.NA,
                "RNE": "o",
            },  # Fail
            {
                "CHILD": "5",
                "DECOM": "31/03/1980",
                "DEC": pd.NA,
                "LS": "C2",
                "REC": pd.NA,
                "RNE": "o",
            },  # Ignore
            {
                "CHILD": "6",
                "DECOM": "31/03/1980",
                "DEC": pd.NA,
                "LS": "C2",
                "REC": pd.NA,
                "RNE": "o",
            },  # Ignore
            {
                "CHILD": "7",
                "DECOM": "31/05/1980",
                "DEC": pd.NA,
                "LS": "C2",
                "REC": pd.NA,
                "RNE": "S",
            },  # Ignore
            {
                "CHILD": "8",
                "DECOM": "31/03/1980",
                "DEC": "30/03/1981",
                "LS": "C2",
                "REC": "E4a",
                "RNE": "o",
            },  # Ignore
            {
                "CHILD": "9",
                "DECOM": "31/03/1980",
                "DEC": pd.NA,
                "LS": "V4",
                "REC": pd.NA,
                "RNE": "o",
            },  # Ignore
        ]
    )

    oc2 = pd.DataFrame(
        {  # 0 Pass, 1 Pass, 2 Pass, 3 Fail (no value), 4 Ignore (age), 5 Ignore (age),
            # 6 Ignore (duration), 7 Ignore (duration), 8 Ignore (LS), 9 Ignore (unmatched)
            "CHILD": ["1", "2", "3", "4", "5", "6", "7", "8", "9", "10"],
            "DOB": [
                "01/04/1970",
                "01/04/1970",
                "01/04/1970",
                "01/04/1970",
                "01/04/1960",  # 20yrs
                "01/04/1978",  # 2yrs
                "01/04/1970",
                "01/04/1970",
                "01/04/1970",
                "01/04/1970",
            ],
            "SDQ_SCORE": [
                "10",
                "1O",
                pd.NA,
                pd.NA,
                pd.NA,
                pd.NA,
                pd.NA,
                pd.NA,
                pd.NA,
                pd.NA,
            ],
            "SDQ_REASON": [
                "SDQ1",
                pd.NA,
                "SDQ2",
                pd.NA,
                pd.NA,
                pd.NA,
                pd.NA,
                pd.NA,
                pd.NA,
                pd.NA,
            ],
        }
    )

    fake_dfs = {"Episodes": eps, "OC2": oc2, "metadata": metadata}

    result = validate(fake_dfs)

    assert result == {
        "OC2": [
            3,
        ]
    }



import pandas as pd

from lac_validator.rule_engine import rule_definition


@rule_definition(
    code="579",
    message="A new decision that the child should be placed for adoption this year cannot start when the previous decision is still open. Decisions to place the child for adoption should also not overlap. The date of any new decision to place the child for adoption must not be before the date placed ceased of previous decisions.",
    affected_fields=["DATE_PLACED", "DATE_PLACED_CEASED"],
    tables=["PlacedAdoption"],
)
def validate(dfs):
    if "PlacedAdoption" not in dfs:
        return {}
    else:
        adopt_placed = dfs["PlacedAdoption"]
        adopt_placed["DATE_PLACED"] = pd.to_datetime(
            adopt_placed["DATE_PLACED"], format="%d/%m/%Y", errors="coerce"
        )
        adopt_placed["DATE_PLACED_CEASED"] = pd.to_datetime(
            adopt_placed["DATE_PLACED_CEASED"], format="%d/%m/%Y", errors="coerce"
        )

        adopt_placed.sort_values(["CHILD", "DATE_PLACED"], inplace=True)

        adopt_placed.reset_index(inplace=True)
        adopt_placed.reset_index(inplace=True)  # Twice on purpose

        adopt_placed["LAG_INDEX"] = adopt_placed["level_0"].shift(-1)

        lag_adopt_placed = adopt_placed.merge(
            adopt_placed,
            how="inner",
            left_on="level_0",
            right_on="LAG_INDEX",
            suffixes=["", "_PREV"],
        )

        # We're only interested in cases where there is more than one row for a child.
        lag_adopt_placed = lag_adopt_placed[
            lag_adopt_placed["CHILD"] == lag_adopt_placed["CHILD_PREV"]
        ]

        # A previous ADOPT_PLACED_CEASED date is null
        mask1 = lag_adopt_placed["DATE_PLACED_CEASED_PREV"].isna()
        # ADOPT_PLACED is before previous ADOPT_PLACED (overlapping dates)
        mask2 = (
            lag_adopt_placed["DATE_PLACED"]
            < lag_adopt_placed["DATE_PLACED_CEASED_PREV"]
        )

        mask = mask1 | mask2

        error_list = lag_adopt_placed["index"][mask].to_list()
        error_list.sort()
        return {"PlacedAdoption": error_list}


def test_validate():
    import pandas as pd

    fake_adopt_place = pd.DataFrame(
        [
            {
                "CHILD": "111",
                "DATE_PLACED": "01/06/2020",
                "DATE_PLACED_CEASED": "05/06/2020",
            },  # 0
            {
                "CHILD": "111",
                "DATE_PLACED": "04/06/2020",
                "DATE_PLACED_CEASED": pd.NA,
            },  # 1 Fails overlaps
            {
                "CHILD": "222",
                "DATE_PLACED": "03/06/2020",
                "DATE_PLACED_CEASED": "04/06/2020",
            },  # 2
            {
                "CHILD": "222",
                "DATE_PLACED": "04/06/2020",
                "DATE_PLACED_CEASED": pd.NA,
            },  # 3
            {
                "CHILD": "222",
                "DATE_PLACED": "07/06/2020",
                "DATE_PLACED_CEASED": "09/06/2020",
            },
            # 4 Fails, previous end is null
            {
                "CHILD": "333",
                "DATE_PLACED": "02/06/2020",
                "DATE_PLACED_CEASED": "04/06/2020",
            },  # 5
            {
                "CHILD": "333",
                "DATE_PLACED": "03/06/2020",
                "DATE_PLACED_CEASED": "09/06/2020",
            },  # 6 Fails overlaps
            {
                "CHILD": "555",
                "DATE_PLACED": pd.NA,
                "DATE_PLACED_CEASED": "05/06/2020",
            },  # 7
            {
                "CHILD": "555",
                "DATE_PLACED": pd.NA,
                "DATE_PLACED_CEASED": "05/06/2020",
            },  # 8
        ]
    )

    fake_dfs = {"PlacedAdoption": fake_adopt_place}

    result = validate(fake_dfs)

    assert result == {"PlacedAdoption": [1, 4, 6]}



import pandas as pd

from lac_validator.rule_engine import rule_definition


@rule_definition(
    code="SW02STG1",
    message="Social worker ID does not begin with the characters ‘SW’ or ‘XX’.",
    affected_fields=["SW_ID"],
    tables=["SWEpisodes"],
)
def validate(dfs):
    if "SWEpisodes" not in dfs:
        return {}
    else:
        df = dfs["SWEpisodes"]

        error_rows = df[~df["SW_ID"].str[:2].isin(["SW", "XX"])].index

        return {"SWEpisodes": error_rows.tolist()}


def test_validate():
    import pandas as pd

    fake_data = pd.DataFrame(
        {
            "SW_ID": ["SW000", "XX000", "aa00", "xx000"],
        }
    )

    fake_dfs = {"SWEpisodes": fake_data}

    result = validate(fake_dfs)

    assert result == {"SWEpisodes": [2, 3]}



import pandas as pd

from lac_validator.rule_engine import rule_definition


@rule_definition(
    code="217t",
    message="Placement type 'P2' is not a valid placement type from 28 October 2023.",
    affected_fields=["PLACE", "DECOM"],
    tables=["Episodes"],
)
def validate(dfs):
    if "Episodes" not in dfs:
        return {}
    else:
        episodes = dfs["Episodes"]
        episodes["DEC"] = pd.to_datetime(
            episodes["DEC"], format="%d/%m/%Y", errors="coerce"
        )
        max_dec_allowed = pd.to_datetime(
            "28/10/2023", format="%d/%m/%Y", errors="coerce"
        )

        mask = episodes["PLACE"].isin(["P2"]) & (
            (episodes["DEC"] > max_dec_allowed) | episodes["DEC"].isna()
        )

        validation_error_mask = mask
        validation_error_locations = episodes.index[validation_error_mask]

        return {"Episodes": validation_error_locations.tolist()}


def test_validate():
    import pandas as pd

    fake_data = pd.DataFrame(
        {
            "PLACE": ["P2", "P2", "P2", "P1"],
            "DEC": [
                "01/11/2023",  # Fail (After code end date)
                "28/10/2023",  # Pass (On code end date)
                pd.NA,  # Fail (Nil end after code end date)
                "01/11/2023",  # Ignore (Not P2)
            ],
        }
    )

    fake_dfs = {"Episodes": fake_data}

    result = validate(fake_dfs)

    assert result == {"Episodes": [0, 2]}



import pandas as pd

from lac_validator.rule_engine import rule_definition


@rule_definition(
    code="578",
    message="The date that the child started to be missing is after the child ceased to be looked after.",
    affected_fields=["MIS_START"],
    tables=["Episodes", "Missing"],
)
def validate(dfs):
    if "Episodes" not in dfs or "Missing" not in dfs:
        return {}
    else:
        episodes = dfs["Episodes"]
        missing = dfs["Missing"]

        episodes["original_index"] = episodes.index

        # convert dates
        episodes["DEC"] = pd.to_datetime(
            episodes["DEC"], format="%d/%m/%Y", errors="coerce"
        )
        episodes["DECOM"] = pd.to_datetime(
            episodes["DECOM"], format="%d/%m/%Y", errors="coerce"
        )
        missing["MIS_START"] = pd.to_datetime(
            missing["MIS_START"], format="%d/%m/%Y", errors="coerce"
        )

        # create period of care blocks
        episodes = episodes.sort_values(["CHILD", "DECOM"])
        episodes = episodes.dropna(subset=["DECOM"])

        episodes["index"] = pd.RangeIndex(0, len(episodes))
        episodes["index+1"] = episodes["index"] + 1
        episodes = episodes.merge(
            episodes,
            left_on="index",
            right_on="index+1",
            how="left",
            suffixes=[None, "_prev"],
        )
        episodes = episodes[
            [
                "original_index",
                "DECOM",
                "DEC",
                "DEC_prev",
                "CHILD",
                "CHILD_prev",
                "LS",
            ]
        ]

        episodes["new_period"] = (episodes["DECOM"] > episodes["DEC_prev"]) | (
            episodes["CHILD"] != episodes["CHILD_prev"]
        )
        episodes["period_id"] = episodes["new_period"].astype(int).cumsum()

        # allocate the same DECOM (min) and DEC (max) to all episodes in a period of care.
        episodes["poc_DECOM"] = episodes.groupby("period_id")["DECOM"].transform("min")
        episodes["poc_DEC"] = episodes.groupby("period_id")["DEC"].transform("max")

        # prepare to merge
        missing["index_ing"] = missing.index

        pocs = pd.DataFrame()
        pocs[["CHILD", "poc_DECOM"]] = episodes.groupby("period_id")[
            ["CHILD", "DECOM"]
        ].first()
        pocs["poc_DEC"] = episodes.groupby("period_id")["DEC"].nth(-1)

        pocs = pocs.merge(missing, on="CHILD", how="right", suffixes=["_eps", "_ing"])
        # If <MIS_START> >=DEC, then no missing/away from placement information should be recorded
        pocs["out_of_poc"] = (pocs["MIS_START"] < pocs["poc_DECOM"]) | (
            (pocs["MIS_START"] > pocs["poc_DEC"]) & pocs["poc_DEC"].notna()
        )

        # Drop rows where child was not present in 'Missing' table.
        pocs = pocs[pocs["index_ing"].notna()]

        mask = pocs.groupby("index_ing")["out_of_poc"].transform("min")
        miss_error_locs = pocs.loc[mask, "index_ing"].astype(int).unique().tolist()

        # In this case it is not necessary to flag the DEC or DECOM that value because that same DEC or DECOM value might have passed with other values of MIS_START. Flagging the DECOM/ DEC value is not specific enough to be helpful to the user.
        return {"Missing": miss_error_locs}


def test_validate():
    import pandas as pd

    fake_data_mis = pd.DataFrame(
        [
            {
                "CHILD": "1",
                "MIS_START": "01/06/2020",
            },  # 0 fail
            {
                "CHILD": "2",
                "MIS_START": "02/06/2020",
            },  # 1 fail
            {
                "CHILD": "3",
                "MIS_START": "03/06/2022",
            },  # 2 fail: fail1 fail2
            {
                "CHILD": "3",
                "MIS_START": "03/06/2020",
            },  # 3 fail with the first, pass with the second : pass
            {
                "CHILD": "4",
                "MIS_START": "04/06/2020",
            },  # 4
            {
                "CHILD": "5",
                "MIS_START": "01/01/1982",
            },  # 5 passes with full period of care DECOM-DEC (7/8) and 9
            {
                "CHILD": "5",
                "MIS_START": "01/01/2000",
            },  # 6 pass: fails with first set of episodes(7/8 poc), but passes with next period of care (9)
            {
                "CHILD": "6",
                "MIS_START": "01/01/1981",
            },  # 7 fail when compared with DECOM
            {
                "CHILD": "77",
                "MIS_START": "01/01/1981",
            },  # 8 pass
            {
                "CHILD": "8888",
                "MIS_START": "01/01/2049",
            },  # 8 pass
        ]
    )
    fake_data_eps = pd.DataFrame(
        [
            {
                "CHILD": "1",
                "DECOM": "01/04/1977",
                "DEC": "31/03/1981",
                "LS": "o",
                "REC": "X1",
            },  # 0
            {
                "CHILD": "2",
                "DECOM": "01/04/1979",
                "DEC": "30/03/1981",
                "LS": "o",
                "REC": pd.NA,
            },  # 1
            {
                "CHILD": "88",
                "DECOM": "01/05/1980",
                "DEC": "01/01/1981",
                "LS": "o",
                "REC": "xx",
            },  # 2
            {
                "CHILD": "3",
                "DECOM": "01/04/1980",
                "DEC": "01/01/1981",
                "LS": "V3",
                "REC": "kk",
            },  # 3
            {
                "CHILD": "3",
                "DECOM": "01/04/2020",
                "DEC": "01/01/2021",
                "LS": "V3",
                "REC": "kk",
            },  # 4 --not poc
            {
                "CHILD": "3",
                "DECOM": "01/01/2021",
                "DEC": "05/10/2021",
                "LS": "V3",
                "REC": "kk",
            },  # 5 due to LS value--
            {
                "CHILD": "4",
                "DECOM": "01/04/1981",
                "DEC": pd.NA,
                "LS": "o",
                "REC": "!!",
            },  # 6
            {
                "CHILD": "5",
                "DECOM": "01/04/1978",
                "DEC": "01/01/1981",
                "LS": "o",
                "REC": "X1",
            },  # 7 --poc start
            {
                "CHILD": "5",
                "DECOM": "01/01/1981",
                "DEC": "01/01/1990",
                "LS": "o",
                "REC": "xx",
            },  # 8  poc end--
            {
                "CHILD": "5",
                "DECOM": "01/04/1983",
                "DEC": "04/06/2020",
                "LS": "o",
                "REC": "kk",
            },  # 9
            {
                "CHILD": "6",
                "DECOM": "01/04/1983",
                "DEC": "04/06/2020",
                "LS": "o",
                "REC": "kk",
            },  # 10
            {
                "CHILD": "8888",
                "DECOM": "01/04/1983",
                "DEC": "04/06/2020",
                "LS": "o",
                "REC": "kk",
            },  # 10
            {
                "CHILD": "8888",
                "DECOM": "04/06/2020",
                "DEC": pd.NA,
                "LS": "o",
                "REC": "kk",
            },  # 10
        ]
    )

    fake_dfs = {"Episodes": fake_data_eps, "Missing": fake_data_mis}

    result = validate(fake_dfs)

    assert result == {"Missing": [0, 1, 2, 7]}



import pandas as pd

from lac_validator.rule_engine import rule_definition


@rule_definition(
    code="529",
    message="Placement provider code of PR3 cannot be associated with placements P1, A3 to A6, K1, K2 and U1 to U6 as these placements cannot be provided by other public organisations.",
    affected_fields=["PLACE", "PLACE_PROVIDER"],
    tables=["Episodes"],
)
def validate(dfs):
    if "Episodes" not in dfs:
        return {}
    else:
        epi = dfs["Episodes"]
        code_list_placement_type = [
            "A3",
            "A4",
            "A5",
            "A6",
            "K1",
            "K2",
            "P1",
            "U1",
            "U2",
            "U3",
            "U4",
            "U5",
            "U6",
        ]
        error_mask = epi["PLACE"].isin(code_list_placement_type) & (
            epi["PLACE_PROVIDER"] == "PR3"
        )
        return {"Episodes": epi.index[error_mask].to_list()}


def test_validate():
    import pandas as pd

    fake_data = pd.DataFrame(
        {
            "PLACE_PROVIDER": ["PR0", "PR1", "PR3", "PR3", pd.NA, "PR3"],
            "PLACE": ["U1", "U2", "U3", "T1", pd.NA, "A3"],
        }
    )

    fake_dfs = {"Episodes": fake_data}

    result = validate(fake_dfs)

    assert result == {"Episodes": [2, 5]}



import pandas as pd

from lac_validator.rule_engine import rule_definition


@rule_definition(
    code="148",
    message="Date episode ceased and reason episode ceased must both be coded, or both left blank.",
    affected_fields=["DEC", "REC"],
    tables=["Episodes"],
)
def validate(dfs):
    if "Episodes" not in dfs:
        return {}
    else:
        df = dfs["Episodes"]

        mask = ((df["DEC"].isna()) & (df["REC"].notna())) | (
            (df["DEC"].notna()) & (df["REC"].isna())
        )

        return {"Episodes": df.index[mask].tolist()}


def test_validate():
    import pandas as pd

    fake_data = pd.DataFrame(
        [
            {
                "CHILD": "111",
                "DECOM": "01/06/2020",
                "DEC": "04/06/2020",
                "REC": pd.NA,
            },  # 0  Fails
            {
                "CHILD": "111",
                "DECOM": "05/06/2020",
                "DEC": "06/06/2020",
                "REC": "X1",
            },  # 1
            {
                "CHILD": "111",
                "DECOM": "06/06/2020",
                "DEC": pd.NA,
                "REC": "X1",
            },  # 2   Fails
            {
                "CHILD": "111",
                "DECOM": "08/06/2020",
                "DEC": "05/06/2020",
                "REC": "X1",
            },  # 3
            {
                "CHILD": "222",
                "DECOM": "05/06/2020",
                "DEC": "06/06/2020",
                "REC": pd.NA,
            },  # 4   Fails
            {
                "CHILD": "333",
                "DECOM": "06/06/2020",
                "DEC": "07/06/2020",
                "REC": "E11",
            },  # 5
            {"CHILD": "333", "DECOM": "07/06/2020", "DEC": pd.NA, "REC": pd.NA},  # 6
            {
                "CHILD": "444",
                "DECOM": "08/06/2020",
                "DEC": "09/06/2020",
                "REC": "X1",
            },  # 7
            {
                "CHILD": "444",
                "DECOM": "09/06/2020",
                "DEC": "10/06/2020",
                "REC": "E11",
            },  # 8
            {"CHILD": "444", "DECOM": "15/06/2020", "DEC": pd.NA, "REC": pd.NA},  # 9
        ]
    )

    fake_dfs = {"Episodes": fake_data}

    result = validate(fake_dfs)

    assert result == {"Episodes": [0, 2, 4]}



import pandas as pd

from lac_validator.rule_engine import rule_definition


@rule_definition(
    code="166",
    message="Date of review is invalid or blank.",
    affected_fields=["REVIEW"],
    tables=["Reviews"],
)
def validate(dfs):
    if "Reviews" not in dfs:
        return {}
    else:
        review = dfs["Reviews"]

        error_mask = pd.to_datetime(
            review["REVIEW"], format="%d/%m/%Y", errors="coerce"
        ).isna()

        validation_error_locations = review.index[error_mask]

        return {"Reviews": validation_error_locations.to_list()}


def test_validate():
    import pandas as pd

    fake_data = pd.DataFrame(
        {
            "REVIEW": ["01/01/2021", "19/02/2010", "38/04/2019", "01/01/19", pd.NA],
        }
    )

    fake_dfs = {"Reviews": fake_data}

    result = validate(fake_dfs)

    assert result == {"Reviews": [2, 3, 4]}



import pandas as pd

from lac_validator.rule_engine import rule_definition


@rule_definition(
    code="367",
    message="The maximum amount of respite care allowable is 75 days in any 12-month period.",
    affected_fields=["LS", "DECOM", "DEC"],
    tables=["Episodes"],
)
def validate(dfs):
    if "Episodes" not in dfs:
        return {}

    episodes = dfs["Episodes"]
    V3_eps = episodes[episodes["LS"] == "V3"]

    V3_eps = V3_eps.dropna(
        subset=["DECOM"]
    )  # missing DECOM should get fixed before looking for this error

    collection_start = pd.to_datetime(
        dfs["metadata"]["collection_start"], format="%d/%m/%Y", errors="coerce"
    )
    collection_end = pd.to_datetime(
        dfs["metadata"]["collection_end"], format="%d/%m/%Y", errors="coerce"
    )
    V3_eps["DECOM_dt"] = pd.to_datetime(
        V3_eps["DECOM"], format="%d/%m/%Y", errors="coerce"
    )
    V3_eps["DEC_dt"] = pd.to_datetime(V3_eps["DEC"], format="%d/%m/%Y", errors="coerce")

    # truncate episode start/end dates to collection start/end respectively
    V3_eps.loc[
        V3_eps["DEC"].isna() | (V3_eps["DEC_dt"] > collection_end), "DEC_dt"
    ] = collection_end
    V3_eps.loc[V3_eps["DECOM_dt"] < collection_start, "DECOM_dt"] = collection_start

    V3_eps["duration"] = (V3_eps["DEC_dt"] - V3_eps["DECOM_dt"]).dt.days
    V3_eps = V3_eps[V3_eps["duration"] > 0]

    V3_eps["year_total_duration"] = V3_eps.groupby("CHILD")["duration"].transform(sum)

    error_mask = V3_eps["year_total_duration"] > 75

    return {"Episodes": V3_eps.index[error_mask].to_list()}


def test_validate():
    import pandas as pd

    test_eps = pd.DataFrame(
        [
            {
                "CHILD": 101,
                "DECOM": "01/01/2000",
                "DEC": "01/06/2000",
                "LS": "V3",
            },  # 0 decom-->col_start
            {
                "CHILD": 2002,
                "DECOM": "01/01/2000",
                "DEC": "01/05/2000",
                "LS": "V3",
            },  # 1 decom-->col_start
            {
                "CHILD": 30003,
                "DECOM": "01/04/2000",
                "DEC": "01/05/2000",
                "LS": "V3",
            },  # 2 ^ Fail!
            {
                "CHILD": 30003,
                "DECOM": "01/05/2000",
                "DEC": "01/07/2000",
                "LS": "xx",
            },  # 3 |
            {
                "CHILD": 30003,
                "DECOM": "01/06/2000",
                "DEC": "01/07/2000",
                "LS": "V3",
            },  # 4 | Fail!
            {
                "CHILD": 30003,
                "DECOM": "01/07/2000",
                "DEC": "01/08/2000",
                "LS": "V3",
            },  # 5 v Fail!
            {
                "CHILD": 400004,
                "DECOM": "01/08/2000",
                "DEC": "01/07/2000",
                "LS": "V3",
            },  # 6
            {
                "CHILD": 400004,
                "DECOM": "01/08/2000",
                "DEC": "01/11/2000",
                "LS": "V3",
            },  # 7 Fail!
            {
                "CHILD": 55555,
                "DECOM": "01/07/2000",
                "DEC": pd.NA,
                "LS": "V3",
            },  # 8 Fail!
        ]
    )

    test_meta = {"collection_start": "01/04/2000", "collection_end": "01/04/2001"}

    test_dfs = {"Episodes": test_eps, "metadata": test_meta}

    result = validate(test_dfs)

    assert result == {"Episodes": [2, 4, 5, 7, 8]}



import pandas as pd

from lac_validator.rule_engine import rule_definition


@rule_definition(
    code="630",
    message="Information on previous permanence option should be returned.",
    affected_fields=["RNE"],
    tables=["Episodes", "PrevPerm"],
)
def validate(dfs):
    if "PrevPerm" not in dfs or "Episodes" not in dfs:
        return {}
    else:
        epi = dfs["Episodes"]
        pre = dfs["PrevPerm"]

        epi["DECOM"] = pd.to_datetime(epi["DECOM"], format="%d/%m/%Y", errors="coerce")
        collection_start = pd.to_datetime(
            dfs["metadata"]["collection_start"], format="%d/%m/%Y", errors="coerce"
        )

        epi = epi.reset_index()

        # Form the episode dataframe which has an 'RNE' of 'S' in this financial year
        epi_has_rne_of_S_in_year = epi[
            (epi["RNE"] == "S") & (epi["DECOM"] >= collection_start)
        ]
        # Merge to see
        # 1) which CHILD ids are missing from the PrevPerm file
        # 2) which CHILD are in the prevPerm file, but don't have the LA_PERM/DATE_PERM field completed where they should be
        # 3) which CHILD are in the PrevPerm file, but don't have the PREV_PERM field completed.
        merged_epi_preperm = epi_has_rne_of_S_in_year.merge(
            pre, on="CHILD", how="left", indicator=True
        )

        error_not_in_preperm = merged_epi_preperm["_merge"] == "left_only"
        error_wrong_values_in_preperm = (merged_epi_preperm["PREV_PERM"] != "Z1") & (
            merged_epi_preperm[["LA_PERM", "DATE_PERM"]].isna().any(axis=1)
        )
        error_null_prev_perm = (merged_epi_preperm["_merge"] == "both") & (
            merged_epi_preperm["PREV_PERM"].isna()
        )

        error_mask = (
            error_not_in_preperm | error_wrong_values_in_preperm | error_null_prev_perm
        )

        error_list = merged_epi_preperm[error_mask]["index"].to_list()
        error_list = list(set(error_list))
        error_list.sort()

        return {"Episodes": error_list}


def test_validate():
    import pandas as pd

    fake_epi = pd.DataFrame(
        {
            "CHILD": ["0", "1", "2", "3", "4", "5", "6", "7", "8"],
            "DECOM": [
                "01/04/2021",
                "01/04/2021",
                "01/04/2021",
                "01/04/2021",
                "31/03/2021",
                "01/04/2021",
                "01/04/2021",
                "01/04/2021",
                "01/04/2021",
            ],
            # 4 will now pass, it's before this year
            "RNE": ["T", "S", "S", "P", "S", "S", "S", "S", "S"],
        }
    )
    fake_pre = pd.DataFrame(
        {
            "CHILD": ["2", "4", "5", "6", "7", "8"],
            "PREV_PERM": ["Z1", "P1", "P1", "Z1", pd.NA, "P1"],
            "LA_PERM": [pd.NA, "352", pd.NA, pd.NA, pd.NA, "352"],
            "DATE_PERM": [pd.NA, pd.NA, "01/05/2000", pd.NA, pd.NA, "05/05/2020"],
        }
    )
    metadata = {"collection_start": "01/04/2021"}

    fake_dfs = {"Episodes": fake_epi, "PrevPerm": fake_pre, "metadata": metadata}

    assert validate(fake_dfs) == {"Episodes": [1, 5, 7]}



import pandas as pd

from lac_validator.rule_engine import rule_definition


@rule_definition(
    code="199",
    message="Episode information shows child has been previously adopted from care. "
    "[NOTE: This only tests the current and previous year data loaded into the tool]",
    affected_fields=["CHILD"],
    tables=["Episodes"],
)
def validate(dfs):
    if "Episodes" not in dfs:
        return {}
    else:
        episodes = dfs["Episodes"]
        episodes["current_year_index"] = episodes.index

        if "Episodes_last" in dfs:
            episodes_last = dfs["Episodes_last"]
            episodes = pd.concat([episodes, episodes_last], axis=0)

    episodes["is_ad"] = episodes["REC"].isin(["E11", "E12"]).astype(int)
    episodes["DECOM"] = pd.to_datetime(
        episodes["DECOM"], format="%d/%m/%Y", errors="coerce"
    )

    episodes = episodes.dropna(subset=["DECOM"]).sort_values("DECOM")

    episodes["ads_to_d8"] = episodes.groupby("CHILD")["is_ad"].cumsum()
    error_mask = (
        episodes["ads_to_d8"] > 0
    ) & ~(  # error if there have been any adoption episodes to date...
        (episodes["ads_to_d8"]) == 1 & episodes["REC"].isin(["E11", "E12"])
    )  # ...unless this is the first

    error_locations = (
        episodes.loc[error_mask, "current_year_index"]
        .dropna()
        .sort_values()
        .astype(int)
        .to_list()
    )

    return {"Episodes": error_locations}


def test_validate():
    import pandas as pd

    fake_data_199_episodes_last = pd.DataFrame(
        [
            {"CHILD": "101", "DECOM": "15/06/2016", "DEC": "20/12/2020", "REC": "E11"},
            {"CHILD": "102", "DECOM": "08/10/2017", "DEC": "03/03/2018", "REC": "E11"},
            {"CHILD": "102", "DECOM": "06/03/2018", "DEC": "12/08/2020", "REC": "E4a"},
            {"CHILD": "103", "DECOM": "11/05/2015", "DEC": "07/04/2019", "REC": "E12"},
            {"CHILD": "104", "DECOM": "26/11/2017", "DEC": "19/07/2020", "REC": "E12"},
            {"CHILD": "107", "DECOM": "26/11/2017", "DEC": pd.NA, "REC": "E12"},
        ]
    )

    fake_data_199_episodes = pd.DataFrame(
        [
            {
                "CHILD": "104",
                "DECOM": "19/07/2020",
                "DEC": "02/03/2021",
                "REC": "E13",
            },  # 0: Pass -> Fail
            {
                "CHILD": "103",
                "DECOM": "16/04/2020",
                "DEC": pd.NA,
                "REC": pd.NA,
            },  # 1: Pass -> Fail
            {
                "CHILD": "103",
                "DECOM": "16/04/2020",
                "DEC": pd.NA,
                "REC": "oo",
            },  # 2: Pass -> Fail
            {
                "CHILD": "105",
                "DECOM": "02/12/2021",
                "DEC": "03/12/2021",
                "REC": "X1",
            },  # 3: Pass
            {
                "CHILD": "105",
                "DECOM": "03/12/2021",
                "DEC": "04/12/2021",
                "REC": "E11",
            },  # 4:Pass
            {
                "CHILD": "105",
                "DECOM": "03/12/2022",
                "DEC": "04/12/2022",
                "REC": "E11",
            },  # 5: Fail
            {
                "CHILD": "105",
                "DECOM": "12/12/2021",
                "DEC": pd.NA,
                "REC": pd.NA,
            },  # 6: Fail
            {
                "CHILD": "107",
                "DECOM": "04/12/2021",
                "DEC": pd.NA,
                "REC": "XXX",
            },  # 7: Fail
            {
                "CHILD": "107",
                "DECOM": "26/11/2017",
                "DEC": "04/12/2021",
                "REC": "E12",
            },  # 8: Pass
        ]
    )

    fake_dfs = {"Episodes": fake_data_199_episodes.copy()}
    result = validate(fake_dfs)
    assert result == {"Episodes": [5, 6, 7]}

    fake_dfs = {
        "Episodes": fake_data_199_episodes,
        "Episodes_last": fake_data_199_episodes_last,
    }
    result = validate(fake_dfs)
    assert result == {"Episodes": [0, 1, 2, 5, 6, 7]}



import pandas as pd

from lac_validator.rule_engine import rule_definition


@rule_definition(
    code="356",
    message="The date the episode ceased is before the date the same episode started.",
    affected_fields=["DECOM"],
    tables=["Episodes"],
)
def validate(dfs):
    if "Episodes" not in dfs:
        return {}
    else:
        episodes = dfs["Episodes"]
        episodes["DECOM"] = pd.to_datetime(
            episodes["DECOM"], format="%d/%m/%Y", errors="coerce"
        )
        episodes["DEC"] = pd.to_datetime(
            episodes["DEC"], format="%d/%m/%Y", errors="coerce"
        )

        error_mask = episodes["DEC"].notna() & (episodes["DEC"] < episodes["DECOM"])

        return {"Episodes": episodes.index[error_mask].to_list()}


def test_validate():
    import pandas as pd

    fake_data = pd.DataFrame(
        {
            "DECOM": ["14/03/2021", "16/06/2019", "03/10/2020", "07/09/2021"],
            "DEC": ["08/12/2020", "24/08/2021", pd.NA, "07/09/2021"],
        }
    )

    fake_dfs = {"Episodes": fake_data}

    result = validate(fake_dfs)

    assert result == {"Episodes": [0]}



import pandas as pd

from lac_validator.rule_engine import rule_definition


@rule_definition(
    code="167",
    message="Data entry for participation is invalid or blank.",
    affected_fields=["REVIEW_CODE"],
    tables=["Reviews"],
)
def validate(dfs):
    if "Reviews" not in dfs:
        return {}

    review = dfs["Reviews"]
    code_list = ["PN0", "PN1", "PN2", "PN3", "PN4", "PN5", "PN6", "PN7"]

    mask = (
        review["REVIEW"].notna() & review["REVIEW_CODE"].isin(code_list)
        | review["REVIEW"].isna() & review["REVIEW_CODE"].isna()
    )

    validation_error_mask = ~mask
    validation_error_locations = review.index[validation_error_mask]

    return {"Reviews": validation_error_locations.tolist()}


def test_validate():
    import pandas as pd

    fake_data = pd.DataFrame(
        {
            "REVIEW": ["01/02/2020", "31/03/2020", "12/12/2019", "05/07/2020", pd.NA],
            "REVIEW_CODE": ["pr0", "child too young", "PN3", "", pd.NA],
        }
    )

    fake_dfs = {"Reviews": fake_data}

    result = validate(fake_dfs)

    assert result == {"Reviews": [0, 1, 3]}



import pandas as pd

from lac_validator.rule_engine import rule_definition


@rule_definition(
    code="222",
    message="Ofsted Unique reference number (URN) should not be recorded for this placement type.",
    affected_fields=["URN"],
    tables=["Episodes"],
)
def validate(dfs):
    # If <URN> provided and <URN> not = ‘XXXXXX’, and where
    # <PL> = ‘P1’; ‘P3’; ‘R1’; ‘R2’; ‘R5’; ‘T0’ ‘T1’; ‘T2’; ‘T3’; ‘T4’ or Z1 then should not be provided
    if "Episodes" not in dfs:
        return {}
    else:
        df = dfs["Episodes"]
        place_code_list = [
            "P1",
            "P3",
            "R1",
            "R2",
            "R5",
            "T0",
            "T1",
            "T2",
            "T3",
            "T4",
            "Z1",
        ]
        mask = (
            (df["PLACE"].isin(place_code_list))
            & (df["URN"].notna())
            & (df["URN"] != "XXXXXXX")
        )
        return {"Episodes": df.index[mask].tolist()}


def test_validate():
    import pandas as pd

    fake_data = pd.DataFrame(
        [
            {"PLACE": "H5", "URN": "XXXXXXX"},  # 0
            {"PLACE": "U1", "URN": "Whatever"},  # 1
            {"PLACE": "U2", "URN": pd.NA},  # 2
            {"PLACE": "T1", "URN": pd.NA},  # 3
            {"PLACE": "R1", "URN": "Whatever"},  # 4  Fail
            {"PLACE": "T2", "URN": "Whatever"},  # 5  Fail
            {"PLACE": "P2", "URN": "Whatever"},  # 6  Pass
        ]
    )

    fake_dfs = {"Episodes": fake_data}

    result = validate(fake_dfs)

    assert result == {"Episodes": [4, 5]}



import pandas as pd

from lac_validator.rule_engine import rule_definition


@rule_definition(
    code="550",
    message="A placement provider code of PR0 can only be associated with placement P1.",
    affected_fields=["PLACE", "PLACE_PROVIDER"],
    tables=["Episodes"],
)
def validate(dfs):
    if "Episodes" not in dfs:
        return {}
    else:
        episodes = dfs["Episodes"]

        mask = (episodes["PLACE"] != "P1") & episodes["PLACE_PROVIDER"].eq("PR0")

        validation_error_locations = episodes.index[mask]
        return {"Episodes": validation_error_locations.tolist()}


def test_validate():
    import pandas as pd

    fake_data = pd.DataFrame(
        {
            "PLACE": ["P1", "P0", "T1", "T3", "T1", "P1", "P0"],
            "PLACE_PROVIDER": ["PR0", "PR2", "PR4", "PR0", pd.NA, "PR0", "PR0"],
        }
    )

    fake_dfs = {"Episodes": fake_data}

    result = validate(fake_dfs)

    assert result == {"Episodes": [3, 6]}



import pandas as pd

from lac_validator.rule_engine import rule_definition


@rule_definition(
    code="552",
    message="Date of Decision to place a child for adoption should be on or prior to the date that the child was placed for adoption.",
    # Field that defines date of decision to place a child for adoption is DATE_PLACED and the start of adoption is defined by DECOM with 'A' placement types.
    affected_fields=["DATE_PLACED", "DECOM"],
    tables=["Episodes", "PlacedAdoption"],
)
def validate(dfs):
    if ("PlacedAdoption" not in dfs) or ("Episodes" not in dfs):
        return {}
    else:
        # get the required datasets
        placed_adoption = dfs["PlacedAdoption"]
        episodes = dfs["Episodes"]
        # keep index values so that they stay the same when needed later on for error locations
        placed_adoption.reset_index(inplace=True)
        episodes.reset_index(inplace=True)

        adoption_eps = episodes[episodes["PLACE"].isin(["A3", "A4", "A5", "A6"])].copy()
        # find most recent adoption decision
        placed_adoption["DATE_PLACED"] = pd.to_datetime(
            placed_adoption["DATE_PLACED"], format="%d/%m/%Y", errors="coerce"
        )
        # remove rows where either of the required values have not been filled.
        placed_adoption = placed_adoption[placed_adoption["DATE_PLACED"].notna()]

        placed_adoption_inds = placed_adoption.groupby("CHILD")["DATE_PLACED"].idxmax(
            skipna=True
        )
        last_decision = placed_adoption.loc[placed_adoption_inds]

        # first time child started adoption
        adoption_eps["DECOM"] = pd.to_datetime(
            adoption_eps["DECOM"], format="%d/%m/%Y", errors="coerce"
        )
        adoption_eps = adoption_eps[adoption_eps["DECOM"].notna()]

        adoption_eps_inds = adoption_eps.groupby("CHILD")["DECOM"].idxmin(skipna=True)
        # full information of first adoption
        first_adoption = adoption_eps.loc[adoption_eps_inds]

        # date of decision and date of start of adoption (DECOM) have to be put in one table
        merged = first_adoption.merge(
            last_decision, on=["CHILD"], how="left", suffixes=["_EP", "_PA"]
        )

        # check to see if date of decision to place is less than or equal to date placed.
        decided_after_placed = merged["DECOM"] < merged["DATE_PLACED"]

        # find the corresponding location of error values per file.
        episode_error_locs = merged.loc[decided_after_placed, "index_EP"]
        placedadoption_error_locs = merged.loc[decided_after_placed, "index_PA"]

        return {
            "PlacedAdoption": placedadoption_error_locs.to_list(),
            "Episodes": episode_error_locs.to_list(),
        }


def test_validate():
    import pandas as pd

    fake_placed_adoption = pd.DataFrame(
        {
            "DATE_PLACED": [
                "08/03/2020",
                "22/06/2020",
                "13/10/2022",
                "24/10/2021",
                pd.NA,
                pd.NA,
                "baddate/2021",
            ],
            "CHILD": ["104", "105", "107", "108", "109", "110", "111"],
        }
    )
    fake_episodes = pd.DataFrame(
        {
            "DECOM": [
                "08/03/2020",
                "22/07/2020",
                "13/10/2021",
                "22/06/2020",
                "26/05/2018",
                pd.NA,
                "01/02/2013",
            ],
            "CHILD": ["104", "105", "107", "108", "109", "110", "111"],
            "PLACE": ["A3", "A4", "A6", "D5", "A3", "A5", "A4"],
        }
    )

    fake_dfs = {"PlacedAdoption": fake_placed_adoption, "Episodes": fake_episodes}
    # get the error function

    result = validate(fake_dfs)
    # check that result of function on provided data is as expected
    assert result == {"PlacedAdoption": [2], "Episodes": [2]}



import pandas as pd

from lac_validator.rule_engine import rule_definition


@rule_definition(
    code="SW08STG2",
    message="A new social worker episode has started, but the previous episode has not ended.",
    affected_fields=["SW_DEC"],
    tables=["SWEpisodes"],
)
def validate(dfs):
    if "SWEpisodes" not in dfs:
        return {}
    else:
        df = dfs["SWEpisodes"]

        df["SW_DECOM"] = pd.to_datetime(
            df["SW_DECOM"], format="%d/%m/%Y", errors="coerce"
        )
        df["SW_DEC"] = pd.to_datetime(df["SW_DEC"], format="%d/%m/%Y", errors="coerce")

        df["SW_DECOM"] = df["SW_DECOM"].fillna(
            "01/01/1901"
        )  # Watch for potential future issues

        df["SW_DECOM"] = df["SW_DECOM"].replace("01/01/1901", pd.NA)

        last_episodes = (
            df.sort_values("SW_DECOM").reset_index().groupby(["CHILD"])["index"].last()
        )
        ended_episodes_df = df.loc[~df.index.isin(last_episodes)]

        ended_episodes_df = ended_episodes_df[
            (ended_episodes_df["SW_DEC"].isna())
            & ended_episodes_df["CHILD"].notna()
            & ended_episodes_df["SW_DECOM"].notna()
        ]
        mask = ended_episodes_df.index.tolist()

        return {"SWEpisodes": mask}


def test_validate():
    import pandas as pd

    fake_data = pd.DataFrame(
        [
            {
                "CHILD": "111",
                "SW_DECOM": "01/06/2020",
                "SW_DEC": "04/06/2020",
            },  # 0
            {
                "CHILD": "111",
                "SW_DECOM": "05/06/2020",
                "SW_DEC": "06/06/2020",
            },  # 1
            {
                "CHILD": "111",
                "SW_DECOM": "06/06/2020",
                "SW_DEC": "08/06/2020",
            },  # 2
            {
                "CHILD": "111",
                "SW_DECOM": "08/06/2020",
                "SW_DEC": "05/06/2020",
            },  # 3
            {
                "CHILD": "222",
                "SW_DECOM": "05/06/2020",
                "SW_DEC": "06/06/2020",
            },  # 4
            {
                "CHILD": "333",
                "SW_DECOM": "06/06/2020",
                "SW_DEC": pd.NA,
            },  # 5   Fails because an episode starts after this one
            {
                "CHILD": "333",
                "SW_DECOM": "07/06/2020",
                "SW_DEC": "08/06/2020",
            },  # 6
            {
                "CHILD": "444",
                "SW_DECOM": "08/06/2020",
                "SW_DEC": "09/06/2020",
            },  # 7
            {
                "CHILD": "444",
                "SW_DECOM": "09/06/2020",
                "SW_DEC": "10/06/2020",
            },  # 8
            {
                "CHILD": "444",
                "SW_DECOM": "15/06/2020",
                "SW_DEC": pd.NA,
            },  # 9
            {
                "CHILD": "555",
                "SW_DECOM": "11/06/2020",
                "SW_DEC": "12/06/2020",
            },  # 10
            {
                "CHILD": "6666",
                "SW_DECOM": "12/06/2020",
                "SW_DEC": "13/06/2020",
            },  # 11
            {
                "CHILD": "6666",
                "SW_DECOM": "13/06/2020",
                "SW_DEC": "14/06/2020",
            },  # 12
            {
                "CHILD": "6666",
                "SW_DECOM": "14/06/2020",
                "SW_DEC": pd.NA,
            },  # 13  Fails because an episode starts after this one
            {
                "CHILD": "6666",
                "SW_DECOM": "15/06/2020",
                "SW_DEC": "16/06/2020",
            },  # 14
            {
                "CHILD": "77777",
                "SW_DECOM": "16/06/2020",
                "SW_DEC": "17/06/2020",
            },  # 15
            {
                "CHILD": "77777",
                "SW_DECOM": "17/06/2020",
                "SW_DEC": "18/06/2020",
            },  # 16
            {
                "CHILD": "77777",
                "SW_DECOM": "18/06/2020",
                "SW_DEC": pd.NA,
            },  # 17
            {
                "CHILD": "999",
                "SW_DECOM": "31/06/2020",
                "SW_DEC": pd.NA,
            },  # 18   Nonsense date, but should pass
            {
                "CHILD": "123",
                "SW_DECOM": pd.NA,
                "SW_DEC": pd.NA,
            },  # 19   Nonsense dates, but should pass
            {
                "CHILD": pd.NA,
                "SW_DECOM": pd.NA,
                "SW_DEC": pd.NA,
            },  # 20   Nonsense, but should pass
        ]
    )

    fake_dfs = {"SWEpisodes": fake_data}

    result = validate(fake_dfs)

    assert result == {"SWEpisodes": [5, 13]}



import pandas as pd

from lac_validator.rule_engine import rule_definition


@rule_definition(
    code="159",
    message="If a child has been recorded as not receiving an intervention for their substance misuse problem, then the additional item on whether an intervention was offered should be completed as well.",
    affected_fields=[
        "SUBSTANCE_MISUSE",
        "INTERVENTION_RECEIVED",
        "INTERVENTION_OFFERED",
    ],
    tables=["OC2"],
)
def validate(dfs):
    if "OC2" not in dfs:
        return {}
    else:
        oc2 = dfs["OC2"]
        mask1 = oc2["SUBSTANCE_MISUSE"].astype(str) == "1"
        mask2 = oc2["INTERVENTION_RECEIVED"].astype(str) == "0"
        mask3 = oc2["INTERVENTION_OFFERED"].isna()

        validation_error = mask1 & mask2 & mask3
        validation_error_locations = oc2.index[validation_error]

        return {"OC2": validation_error_locations.tolist()}


def test_validate():
    import pandas as pd

    fake_data = pd.DataFrame(
        {
            "SUBSTANCE_MISUSE": ["1", "1", "1", "1", 1, "1", pd.NA, pd.NA, 1, 0],
            "INTERVENTION_RECEIVED": ["1", "0", "0", pd.NA, 0, "1", "1", pd.NA, 0, 1],
            "INTERVENTION_OFFERED": [
                pd.NA,
                "0",
                pd.NA,
                pd.NA,
                pd.NA,
                "1",
                pd.NA,
                pd.NA,
                0,
                1,
            ],
        }
    )

    fake_dfs = {"OC2": fake_data}

    result = validate(fake_dfs)

    assert result == {"OC2": [2, 4]}



import pandas as pd

from lac_validator.rule_engine import rule_definition
from lac_validator.fixtures import current_episodes, previous_episodes

import pandas as pd


@rule_definition(
    code="452",
    message="Contradiction between local authority of placement code in the last episode of the previous year and in the first episode of the current year.",
    affected_fields=["PL_LA"],
    tables=["Episodes", "Episodes_last"],
)
def validate(dfs):
    if "Episodes" not in dfs:
        return {}
    if "Episodes_last" not in dfs:
        return {}
    else:
        episodes = dfs["Episodes"]
        episodes_last = dfs["Episodes_last"]

        episodes["DECOM"] = pd.to_datetime(
            episodes["DECOM"], format="%d/%m/%Y", errors="coerce"
        )
        episodes_last["DECOM"] = pd.to_datetime(
            episodes_last["DECOM"], format="%d/%m/%Y", errors="coerce"
        )

        episodes_min = episodes.groupby("CHILD")["DECOM"].idxmin()
        episodes_last_max = episodes_last.groupby("CHILD")["DECOM"].idxmax()

        episodes = episodes[episodes.index.isin(episodes_min)]
        episodes_last = episodes_last[episodes_last.index.isin(episodes_last_max)]

        episodes_merged = (
            episodes.reset_index()
            .merge(
                episodes_last,
                how="left",
                on=["CHILD"],
                suffixes=("", "_last"),
                indicator=True,
            )
            .set_index("index")
        )

        in_both_years = episodes_merged["_merge"] == "both"
        same_rne = episodes_merged["RNE"] == episodes_merged["RNE_last"]
        last_year_open = episodes_merged["DEC_last"].isna()
        different_pl_la = episodes_merged["PL_LA"].astype(str) != episodes_merged[
            "PL_LA_last"
        ].astype(str)

        error_mask = in_both_years & same_rne & last_year_open & different_pl_la

        validation_error_locations = episodes.index[error_mask]

        return {"Episodes": validation_error_locations.tolist()}


def test_validate():
    dfs = {
        "Episodes": current_episodes,
        "Episodes_last": previous_episodes,
    }

    result = validate(dfs)
    assert result == {"Episodes": [4, 6]}



import pandas as pd

from lac_validator.rule_engine import rule_definition


@rule_definition(
    code="193",
    message="Child not identified as having a substance misuse problem but at least one of the two additional items on whether an intervention were offered and received have been completed.",
    affected_fields=[
        "SUBSTANCE_MISUSE",
        "INTERVENTION_RECEIVED",
        "INTERVENTION_OFFERED",
    ],
    tables=["OC2"],
)
def validate(dfs):
    if "OC2" not in dfs:
        return {}
    else:
        oc2 = dfs["OC2"]

        no_substance_misuse = oc2["SUBSTANCE_MISUSE"].isna() | (
            oc2["SUBSTANCE_MISUSE"].astype(str) == "0"
        )
        intervention_not_blank = (
            oc2["INTERVENTION_RECEIVED"].notna() | oc2["INTERVENTION_OFFERED"].notna()
        )

        error_mask = no_substance_misuse & intervention_not_blank
        validation_error_locations = oc2.index[error_mask]

        return {"OC2": validation_error_locations.tolist()}


def test_validate():
    import pandas as pd

    fake_data = pd.DataFrame(
        {
            "SUBSTANCE_MISUSE": [0, pd.NA, 0, "1", pd.NA, pd.NA, 0],
            "INTERVENTION_RECEIVED": [0, pd.NA, "1", 1, "1", pd.NA, pd.NA],
            "INTERVENTION_OFFERED": [0, pd.NA, "1", 1, pd.NA, 0, pd.NA],
        }
    )

    fake_dfs = {"OC2": fake_data}

    result = validate(fake_dfs)

    assert result == {"OC2": [0, 2, 4, 5]}



import pandas as pd

from lac_validator.rule_engine import rule_definition


@rule_definition(
    code="208",
    message="Unique Pupil Number (UPN) for the current year disagrees with the Unique Pupil Number (UPN) already recorded for this child.",
    affected_fields=["UPN"],
    tables=["Header", "Header_last"],
)
def validate(dfs):
    if "Header" not in dfs or "Header_last" not in dfs:
        return {}
    else:
        header = dfs["Header"]
        header_last = dfs["Header_last"]

        header_merged = (
            header.reset_index()
            .merge(
                header_last,
                how="left",
                on=["CHILD"],
                suffixes=("", "_last"),
                indicator=True,
            )
            .set_index("index")
        )

        null_now = header_merged["UPN"].isna()
        null_before = header_merged["UPN_last"].isna()
        in_both_years = header_merged["_merge"] == "both"

        header_merged["UPN"] = header_merged["UPN"].astype(str).str.upper()
        header_merged["UPN_last"] = header_merged["UPN_last"].astype(str).str.upper()
        upn_is_different = (
            (header_merged["UPN"] != header_merged["UPN_last"])
            & ~(null_now & null_before)
            # exclude case where unknown both years null; leave to 442 (missing UPN)
        )

        UN2_to_5 = ["UN2", "UN3", "UN4", "UN5"]
        UN_codes = [
            "UN1",
        ] + UN2_to_5
        valid_unknown_change = (
            (
                header_merged["UPN_last"].eq("UN1") | null_before
            )  # change from UN1/null...
            & header_merged["UPN"].isin(UN2_to_5)
        ) | (  # ...to UN2-5
            null_before & header_merged["UPN_last"].eq("UN1")
        )  # OR, change from null to UN1
        unknown_to_known = (
            header_merged["UPN_last"].isin(UN_codes) | null_before
        ) & ~(  # was either null or an UN-code
            header_merged["UPN"].isin(UN_codes) | null_now
        )  # now neither null nor UN-known

        error_mask = (
            in_both_years & upn_is_different & ~valid_unknown_change & ~unknown_to_known
        )

        error_locations = header.index[error_mask]

        return {"Header": error_locations.to_list()}


def test_validate():
    import pandas as pd

    fake_data_prev = pd.DataFrame(
        {
            "CHILD": [
                "101",
                "102",
                "103",
                "104",
                "105",
                "106",
                "108",
                "109",
                "110",
                "111",
                "33333",
                "44444",
                "1000",
            ],
            "UPN": [
                "UN5",
                "X888888888888",
                "UN1",
                "UN1",
                pd.NA,
                "UN4",
                "UN1",
                pd.NA,
                "a------------",
                "UN2",
                "UN5",
                "H000000000000",
                pd.NA,
            ],
        }
    )
    fake_data = pd.DataFrame(
        {
            "CHILD": [
                "101",
                "102",
                "103",
                "104",
                "105",
                "106",
                "108",
                "109",
                "110",
                "111",
                "55555",
                "66666",
                "1000",
            ],
            "UPN": [
                "H801200001001",
                "O------------",
                "UN1",
                "UN2",
                pd.NA,
                "UN3",
                pd.NA,
                "UN4",
                "A------------",
                "H801200001111",
                "UN5",
                "X999999999999",
                "UN1",
            ],
        }
    )

    fake_dfs = {"Header": fake_data, "Header_last": fake_data_prev}

    result = validate(fake_dfs)

    assert result == {"Header": [1, 5, 6, 12]}
